#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2020  The Project X-Ray Authors.
#
# Use of this source code is governed by a ISC-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/ISC
#
# SPDX-License-Identifier: ISC

import os, re
from prjxray import util


def noprefix(tag):
    n = tag.find('.')
    assert n > 0
    return tag[n + 1:]


def getprefix(tag):
    n = tag.find('.')
    assert n > 0
    return tag[0:n]


def load_pipfile(pipfile, verbose=False):
    '''Returns a set of tags containing real tile type prefixes (ex: INT_L)'''
    todos = set()
    tile_type = None
    with open(pipfile, "r") as f:
        # INT_L.WW2BEG0.SR1BEG_S0
        for line in f:
            tag = line.strip().split(' ')[0]
            prefix_line = getprefix(tag)
            if tile_type is None:
                tile_type = prefix_line
            else:
                assert tile_type == prefix_line
            todos.add(tag)
    return todos, tile_type


def balance_todo_list(
        pipfile,
        todos,
        balance_wire_re,
        balance_wire_direction,
        balance_wire_cnt,
        verbose=False):
    """Balance the contents of the todo list

    Todo list balancing allows to specify the name, direction and minimal number of
    occurrences of a PIP wire in the final todo list.
    The mechanism should be used in cases where a fuzzer times out because of an
    unsolvable todo list, i.e. the netlist and resulting segdata generated from an
    iteration keep segmatch from properly disambiguating the bits for some features.

    When the balance wire name regexp is specified it's guaranteed that all PIPs
    with matching wire name (whether we want to match a src or dst wire has to be
    specified with the --balance-wire-direction switch) will have at least the number
    of entries specified with the --balance-wire-cnt switch in the final todo list.
    """
    orig_todos, tile_type = load_pipfile(pipfile, verbose=verbose)
    if balance_wire_re is not None:
        todo_wires = {}
        verbose and print("Start balancing the TODO list")
        for todo in todos:
            tile_type, dst, src = todo.split(".")
            wire = src
            other_wire = dst
            if balance_wire_direction not in "src":
                wire = dst
                other_wire = src
            balance_wire_match = re.match(balance_wire_re, wire)
            if balance_wire_match is None:
                continue
            if wire not in todo_wires:
                todo_wires[wire] = set()
            todo_wires[wire].add(other_wire)
        for wire, other_wires in todo_wires.items():
            if len(other_wires) >= balance_wire_cnt:
                continue
            else:
                for todo in orig_todos:
                    tile_type, dst, src = todo.split(".")
                    if balance_wire_direction in "src":
                        if wire in src and dst not in todo_wires[wire]:
                            todo_wires[wire].add(dst)
                    else:
                        if wire in dst and src not in todo_wires[wire]:
                            todo_wires[wire].add(src)
                    if len(todo_wires[wire]) == balance_wire_cnt:
                        break
        for wire, other_wires in todo_wires.items():
            if len(other_wires) < balance_wire_cnt:
                verbose and print(
                    "Warning: failed to balance the todo list for wire {}, there are only {} PIPs which meet the requirement: {}"
                    .format(wire, len(other_wires), other_wires))
            for other_wire in other_wires:
                line = tile_type + "."
                if balance_wire_direction in "src":
                    line += other_wire + "." + wire
                else:
                    line += wire + "." + other_wire
                verbose and print("Adding {}".format(line))
                todos.add(line)
        verbose and print("Finished balancing the TODO list")


def maketodo(
        pipfile,
        dbfile,
        intre,
        exclude_re=None,
        balance_wire_re=None,
        balance_wire_direction=None,
        balance_wire_cnt=None,
        not_endswith=None,
        verbose=False):
    '''
    db files start with INT., but pipfile lines start with INT_L
    Normalize by removing before the first dot
    050-intpips doesn't care about contents, but most fuzzers use the tile type prefix
    '''

    todos, tile_type = load_pipfile(pipfile, verbose=verbose)
    verbose and print('%s: %u entries' % (pipfile, len(todos)))
    if not todos:
        verbose and print('%s: %u entries, done!' % (pipfile, len(todos)))
        return

    verbose and print("pipfile todo sample: %s" % list(todos)[0])

    if 0 and verbose:
        print("TODOs")
        for todo in sorted(list(todos)):
            print('  %s' % todo)

    verbose and print('Pre db %s: %u entries' % (dbfile, len(todos)))
    # Allow against empty db
    if os.path.exists(dbfile):
        verbose and print("Loading %s" % dbfile)
        with open(dbfile, "r") as f:
            # INT.BYP_ALT0.BYP_BOUNCE_N3_3 !22_07 !23_07 !25_07 21_07 24_07
            for line in f:
                tag, _bits, mode, _ = util.parse_db_line(line.strip())
                # Only count resolved entries
                if mode:
                    continue
                # INT.BLAH => INT_L.BLAH
                tag = tile_type + '.' + noprefix(tag)

                # bipips works on a subset
                if tag in todos:
                    todos.remove(tag)
                else:
                    verbose and print(
                        "WARNING: couldnt remove %s (line %s)" %
                        (tag, line.strip()))
    else:
        verbose and print("WARNING: dbfile doesnt exist: %s" % dbfile)
    verbose and print('Post db %s: %u entries' % (dbfile, len(todos)))

    drops = 0
    lines = 0
    filtered_todos = set()
    for line in todos:
        include = re.match(intre, line) is not None

        if include and not_endswith is not None:
            include = not line.endswith(not_endswith)

        if include and exclude_re is not None:
            include = re.match(exclude_re, line) is None

        if include:
            filtered_todos.add(line)
        else:
            drops += 1
        lines += 1
    verbose and print('Print %u entries w/ %u drops' % (lines, drops))

    balance_todo_list(
        pipfile, filtered_todos, balance_wire_re, balance_wire_direction,
        balance_wire_cnt, verbose)
    for todo in filtered_todos:
        print(todo)


def run(
        build_dir,
        db_dir,
        pip_dir,
        intre,
        sides,
        l,
        r,
        pip_type,
        seg_type,
        exclude_re=None,
        balance_wire_re=None,
        balance_wire_direction=None,
        balance_wire_cnt=None,
        not_endswith=None,
        verbose=False):
    if db_dir is None:
        db_dir = "%s/%s" % (
            os.getenv("XRAY_DATABASE_DIR"), os.getenv("XRAY_DATABASE"))

    if pip_dir is None:
        pip_dir = "%s/piplist/build/%s" % (
            os.getenv("XRAY_FUZZERS_DIR"), pip_type)

    assert intre, "RE is required"

    for side in sides:
        if side == "l" and not l:
            continue

        if side == "r" and not r:
            continue

        if side == "xl":
            segfile = "l{}".format(seg_type)
            pipfile = "l{}".format(pip_type)
        elif side == "xr":
            segfile = "r{}".format(seg_type)
            pipfile = "r{}".format(pip_type)
        elif side != "":
            segfile = "{}_{}".format(seg_type, side)
            pipfile = "{}_{}".format(pip_type, side)
        else:
            segfile = "{}".format(seg_type)
            pipfile = "{}".format(pip_type)

        maketodo(
            "%s/%s.txt" % (pip_dir, pipfile),
            "%s/segbits_%s.db" % (db_dir, segfile),
            intre,
            exclude_re=exclude_re,
            balance_wire_re=balance_wire_re,
            balance_wire_direction=balance_wire_direction,
            balance_wire_cnt=balance_wire_cnt,
            not_endswith=not_endswith,
            verbose=verbose)


def main():
    import argparse

    parser = argparse.ArgumentParser(
        description="Print list of known but unsolved PIPs")

    parser.add_argument('--verbose', action='store_true', help='')
    parser.add_argument('--build-dir', default="build", help='')
    parser.add_argument('--db-dir', default=None, help='')
    parser.add_argument('--pip-dir', default=None, help='')
    parser.add_argument('--re', required=True, help='')
    parser.add_argument('--exclude-re', required=False, default=None, help='')
    parser.add_argument(
        '--balance-wire-re', required=False, default=None, help='')
    parser.add_argument(
        '--balance-wire-direction', required=False, default="src", help='')
    parser.add_argument(
        '--balance-wire-cnt', required=False, default="1", help='')
    parser.add_argument(
        '--balance-re-wire', required=False, default="src", help='')
    parser.add_argument('--pip-type', default="pips_int", help='')
    parser.add_argument('--seg-type', default="int", help='')
    parser.add_argument('--sides', default="l,r", help='')
    util.add_bool_arg(parser, '--l', default=True, help='')
    util.add_bool_arg(parser, '--r', default=True, help='')
    parser.add_argument(
        '--not-endswith', help='Drop lines if they end with this')
    args = parser.parse_args()

    run(
        build_dir=args.build_dir,
        db_dir=args.db_dir,
        pip_dir=args.pip_dir,
        intre=args.re,
        exclude_re=args.exclude_re,
        balance_wire_re=args.balance_wire_re,
        balance_wire_direction=args.balance_wire_direction,
        balance_wire_cnt=int(args.balance_wire_cnt),
        sides=args.sides.split(','),
        l=args.l,
        r=args.r,
        pip_type=args.pip_type,
        seg_type=args.seg_type,
        not_endswith=args.not_endswith,
        verbose=args.verbose)


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2020  The Project X-Ray Authors.
#
# Use of this source code is governed by a ISC-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/ISC
#
# SPDX-License-Identifier: ISC
from prjxray import segment_map
from prjxray.grid_types import BlockType, GridLoc, GridInfo, BitAlias, Bits, BitsInfo, ClockRegion
from prjxray.tile_segbits_alias import TileSegbitsAlias
import re

CLOCK_REGION_RE = re.compile('X([0-9])Y([0-9])')


class Grid(object):
    """ Object that represents grid for a given database.

  Provides methods to inspect grid by name or location.  Also provides mapping
  of segment offsets for particular grid locations and their tile types.
  """

    def __init__(self, db, tilegrid):
        self.db = db
        self.tilegrid = tilegrid
        self.loc = {}
        self.tileinfo = {}

        clock_regions = {}

        for tile in self.tilegrid:
            tileinfo = self.tilegrid[tile]
            grid_loc = GridLoc(tileinfo['grid_x'], tileinfo['grid_y'])
            assert grid_loc not in self.loc
            self.loc[grid_loc] = tile

            bits = {}

            if 'bits' in tileinfo:
                for k in tileinfo['bits']:
                    segment_type = BlockType(k)
                    base_address = int(tileinfo['bits'][k]['baseaddr'], 0)

                    alias = None
                    if 'alias' in tileinfo['bits'][k]:
                        alias = BitAlias(
                            tile_type=tileinfo['bits'][k]['alias']['type'],
                            start_offset=tileinfo['bits'][k]['alias']
                            ['start_offset'],
                            sites=tileinfo['bits'][k]['alias']['sites'],
                        )

                    bits[segment_type] = Bits(
                        base_address=base_address,
                        frames=tileinfo['bits'][k]['frames'],
                        offset=tileinfo['bits'][k]['offset'],
                        words=tileinfo['bits'][k]['words'],
                        alias=alias,
                    )

            clock_region = None
            if 'clock_region' in tileinfo:
                if tileinfo['clock_region'] is not None:
                    if tileinfo['clock_region'] not in clock_regions:
                        m = CLOCK_REGION_RE.fullmatch(tileinfo['clock_region'])
                        assert m is not None, tileinfo['clock_region']

                        clock_regions[tileinfo['clock_region']] = ClockRegion(
                            name=tileinfo['clock_region'],
                            x=int(m.group(1)),
                            y=int(m.group(2)))

                    clock_region = clock_regions[tileinfo['clock_region']]

            self.tileinfo[tile] = GridInfo(
                bits=bits,
                sites=tileinfo['sites'],
                prohibited_sites=tileinfo['prohibited_sites'],
                tile_type=tileinfo['type'],
                pin_functions=tileinfo.get('pin_functions', {}),
                clock_region=clock_region,
            )

        x, y = zip(*self.loc.keys())
        self._dims = (min(x), max(x), min(y), max(y))

    def tiles(self):
        """ Return list of tiles. """
        return self.tileinfo.keys()

    def tile_locations(self):
        """ Return list of tile locations. """
        return self.loc.keys()

    def dims(self):
        """ Returns (x_min, x_max, y_min, y_max) for given Grid. """
        return self._dims

    def is_populated(self, grid_loc):
        return grid_loc in self.loc

    def loc_of_tilename(self, tilename):
        tileinfo = self.tilegrid[tilename]
        return GridLoc(tileinfo['grid_x'], tileinfo['grid_y'])

    def tilename_at_loc(self, grid_loc):
        return self.loc[grid_loc]

    def gridinfo_at_loc(self, grid_loc):
        return self.tileinfo[self.loc[grid_loc]]

    def gridinfo_at_tilename(self, tilename):
        return self.tileinfo[tilename]

    def iter_all_frames(self):
        for tile, tileinfo in self.tileinfo.items():
            for block_type, bits in tileinfo.bits.items():
                yield BitsInfo(
                    block_type=block_type,
                    tile=tile,
                    bits=bits,
                )

    def get_segment_map(self):
        return segment_map.SegmentMap(self)

    def tile_key(self, tilename):
        gridinfo = self.gridinfo_at_tilename(tilename)
        loc = self.loc_of_tilename(tilename)
        tile_type = gridinfo.tile_type

        return (tile_type, loc.grid_x, -loc.grid_y)

    def get_tile_segbits_at_tilename(self, tilename):
        gridinfo = self.gridinfo_at_tilename(tilename)

        # Check to see if alias is present
        any_alias = False
        for block_type, bits in gridinfo.bits.items():
            if bits.alias is not None:
                any_alias = True

        if any_alias:
            return TileSegbitsAlias(self.db, gridinfo.tile_type, gridinfo.bits)
        else:
            return self.db.get_tile_segbits(gridinfo.tile_type)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2020  The Project X-Ray Authors.
#
# Use of this source code is governed by a ISC-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/ISC
#
# SPDX-License-Identifier: ISC

import os

from prjxray.util import get_part_information, get_part_resources, get_fabric_for_part


def get_environment_variables():
    """
    This function returns a dict of environment settings for a given part.

    It requires a basic setup of the prjxray environment that contains
    information about the databse directory location, the FPGA part, etc.

    One of the settings must be sourced, to set up the basic environment.
    """
    environment = dict()

    # Get base environment variables and resources' paths
    part = os.environ['XRAY_PART']
    database = os.environ['XRAY_DATABASE']
    database_dir = os.environ['XRAY_DATABASE_DIR']
    xray_root = os.environ['XRAY_DIR']
    db_root = os.path.join(database_dir, database)
    res_path = os.path.join(xray_root, 'settings', database)

    # Get device information
    part_info = get_part_information(db_root, part)
    fabric = get_fabric_for_part(db_root, part)
    resources = get_part_resources(res_path, os.environ['XRAY_PART'])

    environment['XRAY_DEVICE'] = part_info['device']
    environment['XRAY_PACKAGE'] = part_info['package']
    environment['XRAY_SPEED_GRADE'] = part_info['speedgrade']
    environment['XRAY_FABRIC'] = fabric
    for number, pin in resources['pins'].items():
        environment['XRAY_PIN_{:02d}'.format(number)] = pin

    return environment


def main():
    # Only dump the environment when the resource.yaml file for the family
    # exists to prevent errors during the creation on the stdout.
    # SKIP_ENV in the environment turns off the environment dump for updating
    # all parts and resources, which will create the resource.yaml file.
    if 'SKIP_ENV' in os.environ:
        return

    environment = get_environment_variables()

    for key, value in environment.items():
        print("export {}={}".format(key, value))


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2022  The Project X-Ray Authors.
#
# Use of this source code is governed by a ISC-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/ISC
#
# SPDX-License-Identifier: ISC
import json
import io
import os
import random
random.seed(int(os.getenv("SEED"), 16))
from prjxray import util
from prjxray import lut_maker
from prjxray import verilog
from prjxray.db import Database


def gen_sites():
    '''
    IOB18S: main IOB of a diff pair
    IOB18M: secondary IOB of a diff pair
    IOB18: not a diff pair. Relatively rare (at least in ROI...2 of them?)
    Focus on IOB18S to start
    '''
    db = Database(util.get_db_root(), util.get_part())
    grid = db.grid()
    for tile_name in sorted(grid.tiles()):
        loc = grid.loc_of_tilename(tile_name)
        gridinfo = grid.gridinfo_at_loc(loc)

        for site_name, site_type in gridinfo.sites.items():
            if site_type in ['IOB18S', 'IOB18M']:
                yield tile_name, site_name


def write_params(params):
    pinstr = 'tile,site,pin,iostandard,drive,slew\n'
    for vals in params:
        pinstr += ','.join(map(str, vals)) + '\n'

    open('params.csv', 'w').write(pinstr)


def use_oserdese2(p, luts, connects):

    p['oddr_mux_config'] = 'none'
    p['tddr_mux_config'] = 'none'

    p['DATA_RATE_OQ'] = verilog.quote(random.choice((
        'SDR',
        'DDR',
    )))

    p['DATA_RATE_TQ'] = verilog.quote(random.choice((
        'BUF',
        'SDR',
        'DDR',
    )))

    if verilog.unquote(p['DATA_RATE_OQ']) == 'SDR':
        data_widths = [2, 3, 4, 5, 6, 7, 8]
    else:
        data_widths = [4, 6, 8]

    p['DATA_WIDTH'] = random.choice(data_widths)

    if p['DATA_WIDTH'] == 4 and verilog.unquote(
            p['DATA_RATE_OQ']) == 'DDR' and verilog.unquote(
                p['DATA_RATE_TQ']) == 'DDR':
        tristate_width = 4
    else:
        tristate_width = 1
    p['SERDES_MODE'] = verilog.quote(random.choice(('MASTER', 'SLAVE')))

    p['TRISTATE_WIDTH'] = tristate_width
    p['OSERDES_MODE'] = verilog.quote(random.choice(('MASTER', 'SLAVE')))

    if p['io']:
        p['TFB'] = '.TFB(tfb_{site}),'.format(**p)
        p['TQ'] = '.TQ({twire}),'.format(**p)
        p['t1net'] = luts.get_next_output_net()
        p['t2net'] = luts.get_next_output_net()
        p['t3net'] = luts.get_next_output_net()
        p['t4net'] = luts.get_next_output_net()
        p['tcenet'] = luts.get_next_output_net()

        for idx in range(4):
            p['IS_T{}_INVERTED'.format(idx + 1)] = random.randint(0, 1)
    else:
        p['TFB'] = '.TFB(),'
        p['TQ'] = '.TQ(),'
        p['t1net'] = ''
        p['t2net'] = ''
        p['t3net'] = ''
        p['t4net'] = ''
        p['tcenet'] = ''

        for idx in range(4):
            p['IS_T{}_INVERTED'.format(idx + 1)] = 0

    p['SRVAL_OQ'] = random.randint(0, 1)
    p['SRVAL_TQ'] = random.randint(0, 1)
    p['INIT_OQ'] = random.randint(0, 1)
    p['INIT_TQ'] = random.randint(0, 1)

    for idx in range(8):
        p['IS_D{}_INVERTED'.format(idx + 1)] = random.randint(0, 1)

    p['IS_CLK_INVERTED'] = random.randint(0, 1)
    p['IS_CLKDIV_INVERTED'] = random.randint(0, 1)

    clk_connections = ''
    p['CLK_USED'] = random.randint(0, 1)
    p['CLKDIV_USED'] = random.randint(0, 1)
    if p['CLK_USED']:
        clk_connections += '''
        .CLK({}),'''.format(luts.get_next_output_net())
    if p['CLKDIV_USED']:
        clk_connections += '''
        .CLKDIV({}),'''.format(luts.get_next_output_net())

    print(
        '''
    (* KEEP, DONT_TOUCH, LOC = "{ologic_loc}" *)
    OSERDESE2 #(
        .SERDES_MODE({OSERDES_MODE}),
        .DATA_RATE_TQ({DATA_RATE_TQ}),
        .DATA_RATE_OQ({DATA_RATE_OQ}),
        .DATA_WIDTH({DATA_WIDTH}),
        .TRISTATE_WIDTH({TRISTATE_WIDTH}),
        .SRVAL_OQ({SRVAL_OQ}),
        .SRVAL_TQ({SRVAL_TQ}),
        .INIT_OQ({INIT_OQ}),
        .INIT_TQ({INIT_TQ}),
        .IS_T1_INVERTED({IS_T1_INVERTED}),
        .IS_T2_INVERTED({IS_T2_INVERTED}),
        .IS_T3_INVERTED({IS_T3_INVERTED}),
        .IS_T4_INVERTED({IS_T4_INVERTED}),
        .IS_D1_INVERTED({IS_D1_INVERTED}),
        .IS_D2_INVERTED({IS_D2_INVERTED}),
        .IS_D3_INVERTED({IS_D3_INVERTED}),
        .IS_D4_INVERTED({IS_D4_INVERTED}),
        .IS_D5_INVERTED({IS_D5_INVERTED}),
        .IS_D6_INVERTED({IS_D6_INVERTED}),
        .IS_D7_INVERTED({IS_D7_INVERTED}),
        .IS_D8_INVERTED({IS_D8_INVERTED}),
        .IS_CLK_INVERTED({IS_CLK_INVERTED}),
        .IS_CLKDIV_INVERTED({IS_CLKDIV_INVERTED})
    ) oserdese2_{site} (
        .OQ({owire}),
        {TFB}
        {TQ}
        {clk_connections}
        .D1({d1net}),
        .D2({d2net}),
        .D3({d3net}),
        .D4({d4net}),
        .D5({d5net}),
        .D6({d6net}),
        .D7({d7net}),
        .D8({d8net}),
        .OCE({ocenet}),
        .RST({rstnet}),
        .T1({t1net}),
        .T2({t2net}),
        .T3({t3net}),
        .T4({t4net}),
        .TCE({tcenet})
        );'''.format(
            clk_connections=clk_connections,
            rstnet=luts.get_next_output_net(),
            d1net=luts.get_next_output_net(),
            d2net=luts.get_next_output_net(),
            d3net=luts.get_next_output_net(),
            d4net=luts.get_next_output_net(),
            d5net=luts.get_next_output_net(),
            d6net=luts.get_next_output_net(),
            d7net=luts.get_next_output_net(),
            d8net=luts.get_next_output_net(),
            ocenet=luts.get_next_output_net(),
            ofb_wire=luts.get_next_input_net(),
            **p),
        file=connects)


def use_direct_and_oddr(p, luts, connects):
    p['oddr_mux_config'] = random.choice((
        'direct',
        'lut',
        'none',
    ))

    if p['io']:
        if p['oddr_mux_config'] != 'lut':
            p['tddr_mux_config'] = random.choice((
                'direct',
                'lut',
                'none',
            ))
        else:
            p['tddr_mux_config'] = random.choice((
                'lut',
                'none',
            ))
    else:
        p['tddr_mux_config'] = 'none'

    # toddr and oddr share the same clk
    if random.randint(0, 1):
        clknet = luts.get_next_output_net()
        p['IS_CLK_INVERTED'] = 0
    else:
        clknet = 'bufg_o'
        p['IS_CLK_INVERTED'] = random.randint(0, 1)

    if p['tddr_mux_config'] == 'direct':
        p['TINIT'] = random.randint(0, 1)
        p['TSRTYPE'] = verilog.quote(random.choice(('SYNC', 'ASYNC')))
        p['TDDR_CLK_EDGE'] = verilog.quote('OPPOSITE_EDGE')

        # Note: it seems that CLK_EDGE setting is ignored for TDDR
        p['TDDR_CLK_EDGE'] = verilog.quote(
            random.choice(('OPPOSITE_EDGE', 'SAME_EDGE')))

        p['t_sr_used'] = random.choice(('None', 'S', 'R'))
        if p['t_sr_used'] == 'None':
            p['t_srnet'] = ''
        elif p['t_sr_used'] == 'S':
            p['srnet'] = luts.get_next_output_net()
            p['t_srnet'] = '.S({}),\n'.format(p['srnet'])
        elif p['t_sr_used'] == 'R':
            p['srnet'] = luts.get_next_output_net()
            p['t_srnet'] = '.R({}),\n'.format(p['srnet'])

        print(
            '''
    (* KEEP, DONT_TOUCH, LOC = "{ologic_loc}" *)
    ODDR #(
        .INIT({TINIT}),
        .SRTYPE({TSRTYPE}),
        .DDR_CLK_EDGE({TDDR_CLK_EDGE}),
        .IS_C_INVERTED({IS_CLK_INVERTED})
    ) toddr_{site} (
        .C({cnet}),
        .D1({d1net}),
        .D2({d2net}),
        .CE({cenet}),
        {t_srnet}
        .Q(tddr_d_{site})
        );
        '''.format(
                cnet=clknet,
                d1net=luts.get_next_output_net(),
                d2net=luts.get_next_output_net(),
                cenet=luts.get_next_output_net(),
                **p),
            file=connects)

    if p['tddr_mux_config'] == 'direct':
        print(
            '''
    assign {twire} = tddr_d_{site};'''.format(**p, ),
            file=connects)
    elif p['tddr_mux_config'] == 'lut':
        print(
            '''
    assign {twire} = {lut};'''.format(lut=luts.get_next_output_net(), **p),
            file=connects)
        pass
    elif p['tddr_mux_config'] == 'none':
        pass
    else:
        assert False, p['tddr_mux_config']

    if p['oddr_mux_config'] == 'direct':
        p['QINIT'] = random.randint(0, 1)
        p['SRTYPE'] = verilog.quote(random.choice(('SYNC', 'ASYNC')))
        p['ODDR_CLK_EDGE'] = verilog.quote(
            random.choice((
                'OPPOSITE_EDGE',
                'SAME_EDGE',
            )))

        p['o_sr_used'] = random.choice(('None', 'S', 'R'))
        if p['o_sr_used'] == 'None':
            p['o_srnet'] = ''
        elif p['o_sr_used'] == 'S':
            if 'srnet' not in p:
                p['srnet'] = luts.get_next_output_net()
            p['o_srnet'] = '.S({}),\n'.format(p['srnet'])
        elif p['o_sr_used'] == 'R':
            if 'srnet' not in p:
                p['srnet'] = luts.get_next_output_net()
            p['o_srnet'] = '.R({}),\n'.format(p['srnet'])

        print(
            '''
    (* KEEP, DONT_TOUCH, LOC = "{ologic_loc}" *)
    ODDR #(
        .INIT({QINIT}),
        .SRTYPE({SRTYPE}),
        .DDR_CLK_EDGE({ODDR_CLK_EDGE}),
        .IS_C_INVERTED({IS_CLK_INVERTED})
    ) oddr_{site} (
        .C({cnet}),
        .D1({d1net}),
        .D2({d2net}),
        .CE({cenet}),
        {o_srnet}
        .Q(oddr_d_{site})
        );
        '''.format(
                cnet=clknet,
                d1net=luts.get_next_output_net(),
                d2net=luts.get_next_output_net(),
                cenet=luts.get_next_output_net(),
                **p),
            file=connects)

    if p['oddr_mux_config'] == 'direct':
        print(
            '''
    assign {owire} = oddr_d_{site};'''.format(**p, ),
            file=connects)
    elif p['oddr_mux_config'] == 'lut':
        print(
            '''
    assign {owire} = {lut};'''.format(lut=luts.get_next_output_net(), **p),
            file=connects)
        pass
    elif p['oddr_mux_config'] == 'none':
        pass
    else:
        assert False, p['oddr_mux_config']


def run():
    iostandards = [
        'LVCMOS12', 'LVCMOS15', 'LVCMOS18'
    ]
    iostandard = random.choice(iostandards)

    if iostandard in ['LVCMOS12']:
        drives = [2, 4, 6, 8]
    elif iostandard in ['LVCMOS15', 'LVCMOS18']:
        drives = [2, 4, 6, 8, 12, 16]
    else:
        assert False, "This should be unreachable"

    slews = ['FAST', 'SLOW']
    pulls = ["NONE", "KEEPER", "PULLDOWN", "PULLUP"]

    luts = lut_maker.LutMaker()

    connects = io.StringIO()

    tile_params = []
    params = []

    ndio = 0
    ndo = 0
    for idx, (tile, site) in enumerate(gen_sites()):
        if idx == 0:
            continue

        p = {}
        p['tile'] = tile
        p['site'] = site
        p['ilogic_loc'] = site.replace('IOB', 'ILOGIC')
        p['ologic_loc'] = site.replace('IOB', 'OLOGIC')
        p['IOSTANDARD'] = verilog.quote(iostandard)
        p['PULLTYPE'] = verilog.quote(random.choice(pulls))
        p['DRIVE'] = random.choice(drives)
        p['SLEW'] = verilog.quote(random.choice(slews))

        p['io'] = random.randint(0, 1)
        p['owire'] = 'do_buf[{}]'.format(idx - 1)

        if p['io']:
            p['pad_wire'] = 'dio[{}]'.format(ndio)
            ndio += 1

            p['iwire'] = 'di_buf[{}]'.format(idx - 1)
            p['twire'] = 't[{}]'.format(idx - 1)
        else:
            p['pad_wire'] = 'do[{}]'.format(ndo)
            ndo += 1

        params.append(p)
        tile_params.append(
            (
                tile, site, p['pad_wire'], iostandard, p['DRIVE'],
                verilog.unquote(p['SLEW']) if p['SLEW'] else None,
                verilog.unquote(p['PULLTYPE'])))

    write_params(tile_params)

    print(
        '''
`define N_DO {n_do}
`define N_DIO {n_dio}

module top(input clk, output wire [`N_DO-1:0] do, inout wire [`N_DIO-1:0] dio);
    wire [(`N_DIO+`N_DO)-1:0] di_buf;
    wire [(`N_DIO+`N_DO)-1:0] do_buf;
    wire [(`N_DIO+`N_DO)-1:0] t;
    '''.format(n_dio=ndio, n_do=ndo))

    # Always output a LUT6 to make placer happy.
    print(
        '''
        (* KEEP, DONT_TOUCH *)
        LUT6 dummy_lut();

        wire bufg_o;
        (* KEEP, DONT_TOUCH *)
        BUFG (.O(bufg_o));
        ''')

    for p in params:
        if p['io']:
            print(
                '''
        wire oddr_d_{site};

        (* KEEP, DONT_TOUCH, LOC = "{site}" *)
        IOBUF #(
            .IOSTANDARD({IOSTANDARD})
        ) obuf_{site} (
            .IO({pad_wire}),
            .I({owire}),
            .O({iwire}),
            .T({twire})
            );
            '''.format(**p),
                file=connects)
        else:
            print(
                '''
        wire oddr_d_{site};

        (* KEEP, DONT_TOUCH, LOC = "{site}" *)
        OBUF #(
            .IOSTANDARD({IOSTANDARD})
        ) obuf_{site} (
            .O({pad_wire}),
            .I({owire})
            );
            '''.format(**p),
                file=connects)

        p['use_oserdese2'] = random.randint(0, 1)
        if p['use_oserdese2']:
            use_oserdese2(p, luts, connects)
        else:
            use_direct_and_oddr(p, luts, connects)

    for l in luts.create_wires_and_luts():
        print(l)

    print(connects.getvalue())

    print("endmodule")

    with open('params.jl', 'w') as f:
        json.dump(params, f, indent=2)


if __name__ == '__main__':
    run()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2020  The Project X-Ray Authors.
#
# Use of this source code is governed by a ISC-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/ISC
#
# SPDX-License-Identifier: ISC

import json
from prjxray import bitstream


def gen_addrs():
    for block_type, top_bottom, cfg_row, cfg_col, frame_count in bitstream.gen_part_base_addrs(
    ):
        yield bitstream.addr_bits2word(
            block_type, top_bottom, cfg_row, cfg_col, 0), frame_count


def run(verbose=False):
    for addr, frame_count in sorted(gen_addrs()):
        print("0x%08X: %u" % (addr, frame_count))


def main():
    import argparse

    parser = argparse.ArgumentParser(
        description='Print number of frames at a base address')
    args = parser.parse_args()

    run(verbose=False)


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2020  The Project X-Ray Authors.
#
# Use of this source code is governed by a ISC-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/ISC
#
# SPDX-License-Identifier: ISC

import re
import argparse
import json
import functools

NUMBER_RE = re.compile(r'\d+$')


def check_sequential(speed_model):
    # combinational path models do not contain
    # the following keywords
    timing_keywords = {
        'setup': 'setup',
        'remov': 'removal',
        'hold': 'hold',
        'recov': 'recovery',
        'removal': 'removal',
        'recovery': 'recovery'
    }
    tmp = speed_model.split('_')
    for keyword in sorted(timing_keywords):
        if keyword in tmp:
            # return found keyword and it's map in SDF
            return [keyword, timing_keywords[keyword]]

    return None


# FF's can be configured to work as FF or latch
def check_ff_latch(speed_model):

    tmp = speed_model.split('_')
    if 'ff' in tmp:
        return 'ff'
    elif 'lat' in tmp:
        return 'lat'
    else:
        return None


# some bels have duplicate names e.g bufmrce_bufmrce
# this function cleans them
def clean_bname(bname):
    tmp = bname.split('_')
    if len(tmp) > 1 and tmp[0] == tmp[1]:
        return '_'.join(tmp[1:])
    return bname


def find_aliased_pin(pin, model, pin_aliases):
    """
    Searches for aliased pins in the timing model.
    The check is done using data from pin_aliases dictionary.
    The dictionary has an entry for each aliased pin.
    Each entry has two fields:

    * names : a list of all the possible aliases
    * is_property_related: a flag saying if the alias is in fact
            pin name combined with BEL property (e.g. Q[LH] pins
            in FF - in this case the pin name is Q [original name],
            but is named Q[LH] in the timing model. The suffix
            determines polarity of the FF's set/reset input).

    If is_property_related is set the function returns the original
    pin name, aliased name is returned otherwise.

    Parameters
    ----------
        pin: str
            Pin name to look for
        model: str
            Timing model
        pin_aliases: dict
            A dict of list of aliases for given bel/site

    Returns
    -------
        bool, str

        The first bool value is set to true if pin is found
        in the timing model, false otherwise.

        The second returned value is found pin name. If pin
        is not found None is returned

    >>> find_aliased_pin("a", "a_b_some_test_string", None)
    (False, None)

    >>> find_aliased_pin("d", "din_dout_setup", {"D": {"names" : ["din"], "is_property_related" : False}})
    (True, 'din')

    >>> find_aliased_pin("d", "din_dout_setup", {"D": {"names" : ["din"], "is_property_related" : True}})
    (True, 'd')

    >>> find_aliased_pin("d", "din_dout_setup", {"D": {"names" : ["notdin"], "is_property_related" : True}})
    (False, None)
    """
    if (pin_aliases is not None) and (pin.upper() in pin_aliases):
        for alias in pin_aliases[pin.upper()]['names']:
            single_word_alias = (len(alias.split('_')) == 1)
            pin_alias = alias.lower()
            if single_word_alias:
                model_to_check = model.split('_')
            else:
                model_to_check = model
            if pin_alias in model_to_check:
                if pin_aliases[pin.upper()]['is_property_related']:
                    return True, pin.lower()
                else:
                    return True, pin_alias

    return False, None


def instance_in_model(instance, model):

    if len(instance.split('_')) == 1:
        # instance name is one word, search it in the model
        return instance in model.split('_')
    else:
        # instance name is multi word, search for a string
        return instance in model


def create_pin_in_model(pin_aliases):
    """
    Checks if a given pin belongs to the model.

    Parameters
    ----------
        pin: str
            Pin name to look for
        pin_aliases: dict
            A dict of list of aliases for given bel/site
        model: str
            Timing model name
        direction: str
            Optional pin direction suffix [IN|OUT]

    Returns
    -------
        bool, str

        The first returned value is set to true if pin is found,
        false otherwise.

        The second returned value contains found pin name. If the
        pin is not found, None is returned.

    >>> create_pin_in_model(None)("d", "ff_init_din_q", "in")
    (True, 'din')

    >>> create_pin_in_model(None)("q", "ff_init_clk_q", None)
    (True, 'q')

    >>> create_pin_in_model({"Q": {"names" : ["QL", "QH"], "is_property_related" : True}})("q", "ff_init_clk_ql", None)
    (True, 'q')

    >>> create_pin_in_model(None)("logic_out", "my_cell_i_logic_out", None)
    (True, 'logic_out')

    >>> create_pin_in_model({"LOGIC_OUT": {"names" : ["LOGIC_O", "O"], "is_property_related" : False}})("logic_out", "my_cell_i_logic_o", None)
    (True, 'logic_o')

    >>> create_pin_in_model({"LOGIC_OUT": {"names" : ["LOGIC_O", "O"], "is_property_related" : False}})("logic_out", "my_cell_i_o", None)
    (True, 'o')
    """

    @functools.lru_cache(maxsize=10000)
    def pin_in_model(pin, model, direction=None):
        # strip site location
        model = model.split(':')[0]

        extended_pin_name = pin
        aliased_pin, aliased_pin_name = find_aliased_pin(
            pin.upper(), model, pin_aliases)

        # some timings reports pins with their directions
        # this happens for e.g. CLB reg_init D pin, which
        # timing is reported as DIN
        if direction is not None:
            extended_pin_name = pin + direction

        if instance_in_model(pin, model):
            return True, pin
        elif instance_in_model(extended_pin_name, model):
            return True, extended_pin_name
        elif aliased_pin:
            return True, aliased_pin_name
        else:
            return False, None

    return pin_in_model


def remove_pin_from_model(pin, model):
    """
    Removes the pin from model name if present.

    Arguments
    ---------

    pin: str
        Pin name
    mode: str
        Timing model name

    Returns
    -------

    str
        Updated timing model name

    >>> remove_pin_from_model("q", "ff_init_d_q")
    'ff_init_d'
    >>> remove_pin_from_model("q", "ff_init_d_ql")
    'ff_init_d_ql'
    >>> remove_pin_from_model("logic_out", "ff_init_d_logic_out")
    'ff_init_d'
    >>> remove_pin_from_model("logic_out", "ff_init_d_second_out")
    'ff_init_d_second_out'
    """

    if len(pin.split('_')) == 1:
        # pin name is one word, search it in the model
        tmp = model.split('_')
        if pin in tmp:
            tmp.remove(pin)
            return "_".join(tmp)
        else:
            return model
    else:
        # pin name is multi word, search for a string
        return "_".join(list(filter(None, model.replace(pin, '').split('_'))))


def merged_dict(itr):
    """ Create a merged dict of dict (of dict) based on input.

    Input is an iteratable of (keys, value).

    Return value is root dictionary

    Keys are successive dictionaries indicies.  For example:
    (('a', 'b', 'c'), 1)

    would set:

    output['a']['b']['c'] = 1

    This function returns an error if two values conflict.

    >>> merged_dict(((('a', 'b', 'c'), 1), (('a', 'b', 'd'), 2)))
    {'a': {'b': {'c': 1, 'd': 2}}}

    """

    output = {}
    for keys, value in itr:
        target = output
        for key in keys[:-1]:
            if key not in target:
                target[key] = {}
            target = target[key]

        if keys[-1] in target:
            assert target[keys[-1]] == value, (keys, value, target[keys[-1]])
        else:
            target[keys[-1]] = value

    return output


def extract_properties(tile, site, bel, properties, model):

    if tile not in properties:
        return None
    if site not in properties[tile]:
        return None
    if bel not in properties[tile][site]:
        return None

    model_properties = dict()

    for prop in properties[tile][site][bel]:
        if prop in model_properties:
            continue
        if instance_in_model(prop.lower(), model):
            # if there is property there must be value
            # value always follow the property
            for value in properties[tile][site][bel][prop]:
                value = value.replace(',', '')
                prop_val_str = "_".join([prop, value])
                if instance_in_model(prop_val_str.lower(), model):
                    model_properties[prop] = value
                    break

    return model_properties


def parse_raw_timing(fin):
    with open(fin, "r") as f:
        for line in f:
            raw_data = line.split()
            slice = raw_data[0]

            sites_count = int(raw_data[1])
            loc = 2
            for site in range(0, sites_count):

                site_name = raw_data[loc]
                bels_count = int(raw_data[loc + 1])

                # read all BELs data within
                loc += 2
                for bel in range(0, bels_count):
                    bel = raw_data[loc]
                    delay_count = int(raw_data[loc + 1])

                    # get all the delays
                    loc += 2
                    for delay in range(0, delay_count):
                        speed_model = raw_data[loc]

                        # each timing entry reports 5 delays
                        timing = [
                            raw_data[d + 1 + loc].split(':')
                            for d in range(0, 5)
                        ]

                        yield slice, site_name, bel, speed_model, timing

                        # 5 delay values + name
                        loc += 6


def read_raw_timings(fin, properties, pins, site_pins, pin_alias_map):
    def inner():
        raw = list(parse_raw_timing(fin))

        pin_in_models = {}

        for slice, site_name, bel, speed_model, timing in raw:
            btype = bel.lower()
            delay_btype = clean_bname(btype)
            delay_btype_orig = delay_btype

            # all the bel names seem to start with "bel_d_"
            # let's get rid of it
            if speed_model.startswith('bel_d_'):
                speed_model = speed_model[6:]

            # keep original speed model string to use as unique dict entry
            speed_model_orig = speed_model

            # if more than one BEL type exists in the slice
            # location is added at the end of the name
            tmp = speed_model.split(':')
            speed_model = tmp[0]

            bel_location = site_name
            if len(tmp) > 2:
                bel_location += "/" + "/".join(tmp[2:])

            bel_location = bel_location.upper()

            sequential = check_sequential(speed_model)
            if sequential is not None:
                tmp = speed_model.split('_')
                tmp.remove(sequential[0])
                speed_model = '_'.join(tmp)

            bel_input = None
            bel_output = None
            bel_clock = None

            # strip btype from speed model so we can search for pins
            speed_model_clean = speed_model
            if speed_model.startswith(delay_btype):
                speed_model_clean = speed_model[len(delay_btype):]

            # remove properties from the model
            speed_model_properties = extract_properties(
                slice, site_name, delay_btype_orig, properties,
                speed_model_clean)
            if speed_model_properties is not None:
                for prop in speed_model_properties:
                    # properties values in the model always follow properties name
                    prop_string = "_".join(
                        [prop, speed_model_properties[prop]])
                    speed_model_clean = remove_pin_from_model(
                        prop_string.lower(), speed_model_clean)

            # Get pin alias map
            if delay_btype not in pin_in_models:
                pin_aliases = pin_alias_map.get(delay_btype, None)
                pin_in_models[delay_btype] = create_pin_in_model(pin_aliases)

            pin_in_model = pin_in_models[delay_btype]

            # locate pins
            for pin in pins[slice][site_name][delay_btype_orig]:
                orig_pin = pin
                pim, pin = pin_in_model(pin.lower(), speed_model_clean, 'in')

                if pim:
                    if pins[slice][site_name][delay_btype_orig][orig_pin][
                            'is_clock'] and not pins[slice][site_name][
                                delay_btype_orig][orig_pin]['is_part_of_bus']:
                        bel_clock = pin
                        bel_clock_orig_pin = orig_pin
                    elif pins[slice][site_name][delay_btype_orig][orig_pin][
                            'direction'] == 'IN':
                        bel_input = pin
                    elif pins[slice][site_name][delay_btype_orig][orig_pin][
                            'direction'] == 'OUT':
                        bel_output = pin
                    speed_model_clean = remove_pin_from_model(
                        pin.lower(), speed_model_clean)

            # Some speed models describe delays from/to site pins instead of BEL pins
            if bel_clock is None:
                for pin in site_pins[slice][site_name.lower()]:
                    orig_pin = pin
                    pim, pin = pin_in_model(pin.lower(), speed_model_clean)
                    if pim:
                        if site_pins[slice][site_name.lower(
                        )][orig_pin]['is_clock'] and not site_pins[slice][
                                site_name.lower()][orig_pin]['is_part_of_bus']:
                            bel_clock = pin
                            bel_clock_orig_pin = orig_pin
                            speed_model_clean = remove_pin_from_model(
                                pin.lower(), speed_model_clean)

            if bel_input is None:
                # search site inputs
                for pin in site_pins[slice][site_name.lower()]:
                    orig_pin = pin
                    pim, pin = pin_in_model(
                        pin.lower(), speed_model_clean, 'in')
                    if pim:
                        if site_pins[slice][site_name.lower(
                        )][orig_pin]['direction'] == 'IN':
                            bel_input = pin
                            speed_model_clean = remove_pin_from_model(
                                pin.lower(), speed_model_clean)

            if bel_output is None:
                for pin in site_pins[slice][site_name.lower()]:
                    orig_pin = pin
                    pim, pin = pin_in_model(pin.lower(), speed_model_clean)
                    if pim:
                        if site_pins[slice][site_name.lower(
                        )][orig_pin]['direction'] == 'OUT':
                            bel_output = pin
                            speed_model_clean = remove_pin_from_model(
                                pin.lower(), speed_model_clean)

            # if we couldn't find input, check if the clock is the
            # only input. This applies only to combinational paths
            if (sequential is None) and (bel_input is None) and (bel_clock is
                                                                 not None):
                if bel_clock_orig_pin in site_pins[slice][site_name.lower()] and \
                site_pins[slice][site_name.lower(
                )][bel_clock_orig_pin]['direction'] == 'IN':
                    bel_input = bel_clock

            # if we still don't have the input check if the input
            # is wider than 1 bit and timing defined for the whole
            # port
            if (bel_input is None) or (bel_output is None):
                for pin in pins[slice][site_name][delay_btype_orig]:
                    number = NUMBER_RE.search(pin)
                    if number is not None:
                        orig_pin = pin[:-(len(str(number.group())))]
                        orig_pin_full = pin
                        pim, pin = pin_in_model(
                            orig_pin.lower(), speed_model_clean)
                        if not pim:
                            # some inputs pins are named with unsignificant zeros
                            # remove ti and try again
                            orig_pin = orig_pin + str(int(number.group()))
                            pim, pin = pin_in_model(
                                orig_pin.lower(), speed_model_clean)

                        if pim:
                            if pins[slice][site_name][delay_btype_orig][orig_pin_full]['direction'] == 'IN' \
                                and bel_input is None:
                                bel_input = pin
                            if pins[slice][site_name][delay_btype_orig][orig_pin_full]['direction'] == 'OUT' \
                                and bel_output is None:
                                bel_output = pin
                            speed_model_clean = remove_pin_from_model(
                                orig_pin.lower(), speed_model_clean)

            # check if the input is not a BEL property
            if bel_input is None:
                # if there is anything not yet decoded
                if len(speed_model_clean.split("_")) > 1:
                    if len(speed_model_properties.keys()) == 1:
                        bel_input = list(speed_model_properties.keys())[0]

            # if we still don't have input, give up
            if bel_input is None:
                continue

            # restore speed model name
            speed_model = delay_btype + speed_model_clean

            if sequential is not None:
                if bel_clock is None:
                    continue

                if bel_output is None and bel_clock is None or \
                    bel_output is None and bel_clock == bel_input:
                    continue
            else:
                if bel_input is None or bel_output is None:
                    continue

            delay_btype = speed_model
            # add properties to the delay_btype
            if speed_model_properties is not None:
                for prop in sorted(speed_model_properties):
                    prop_string = "_".join(
                        [prop, speed_model_properties[prop]])
                    delay_btype += "_" + prop_string

            yield (slice, bel_location, delay_btype, speed_model_orig,
                   'type'), btype.upper()
            yield (
                slice, bel_location, delay_btype, speed_model_orig,
                'input'), bel_input.upper()

            if bel_output is not None:
                yield (
                    slice, bel_location, delay_btype, speed_model_orig,
                    'output'), bel_output.upper()

            if bel_clock is not None:
                yield (
                    slice, bel_location, delay_btype, speed_model_orig,
                    'clock'), bel_clock.upper()

            yield (
                slice, bel_location, delay_btype, speed_model_orig,
                'location'), bel_location.upper()

            #XXX: debug
            yield (
                slice, bel_location, delay_btype, speed_model_orig,
                'model'), speed_model_orig

            if sequential is not None:
                assert bel_clock is not None, (
                    slice, bel_location, delay_btype, speed_model_orig)
                yield (
                    slice, bel_location, delay_btype, speed_model_orig,
                    'sequential'), sequential[1]

            for t, v in timing:
                yield (
                    slice, bel_location, delay_btype, speed_model_orig, t), v

    return merged_dict(inner())


def read_bel_properties(properties_file, properties_map):
    def inner():
        with open(properties_file, 'r') as f:
            for line in f:
                raw_props = line.split()
                tile = raw_props[0]
                sites_count = int(raw_props[1])
                prop_loc = 2

                if sites_count == 0:
                    yield (tile, ), {}

                for site in range(0, sites_count):
                    site_name = raw_props[prop_loc]
                    bels_count = int(raw_props[prop_loc + 1])
                    prop_loc += 2

                    for bel in range(0, bels_count):
                        bel_name = raw_props[prop_loc]
                        bel_name = clean_bname(bel_name)
                        bel_name = bel_name.lower()
                        bel_properties_count = int(raw_props[prop_loc + 1])

                        props = 0
                        prop_loc += 2
                        for prop in range(0, bel_properties_count):
                            prop_name = raw_props[prop_loc]

                            # the name always starts with "CONFIG." and ends with ".VALUES"
                            # let's get rid of that
                            if prop_name.startswith(
                                    'CONFIG.') and prop_name.endswith(
                                        '.VALUES'):
                                prop_name = prop_name[7:-7]

                            prop_values_count = int(raw_props[prop_loc + 1])

                            if prop_name not in [
                                    'RAM_MODE',
                                    'WRITE_WIDTH_A',
                                    'WRITE_WIDTH_B',
                                    'READ_WIDTH_A',
                                    'READ_WIDTH_B',
                            ]:
                                if bel_name in properties_map:
                                    if prop_name in properties_map[bel_name]:
                                        prop_name = properties_map[bel_name][
                                            prop_name]

                                yield (tile, site_name, bel_name, prop_name), \
                                        raw_props[prop_loc + 2:prop_loc + 2 +
                                                        prop_values_count]
                                props += 1

                            prop_loc += 2 + prop_values_count

                        if props == 0:
                            yield (tile, site_name, bel_name), {}

    return merged_dict(inner())


def read_bel_pins(pins_file):
    def inner():
        with open(pins_file, 'r') as f:
            for line in f:
                raw_pins = line.split()
                tile = raw_pins[0]
                sites_count = int(raw_pins[1])
                pin_loc = 2

                if sites_count == 0:
                    yield (tile, ), {}

                for site in range(0, sites_count):
                    site_name = raw_pins[pin_loc]
                    bels_count = int(raw_pins[pin_loc + 1])
                    pin_loc += 2

                    for bel in range(0, bels_count):
                        bel_name = raw_pins[pin_loc]
                        bel_name = clean_bname(bel_name)
                        bel_name = bel_name.lower()
                        bel_pins_count = int(raw_pins[pin_loc + 1])

                        pin_loc += 2
                        for pin in range(0, bel_pins_count):
                            pin_name = raw_pins[pin_loc]
                            pin_direction = raw_pins[pin_loc + 1]
                            pin_is_clock = raw_pins[pin_loc + 2]
                            pin_is_part_of_bus = raw_pins[pin_loc + 3]

                            yield (
                                tile, site_name, bel_name, pin_name,
                                'direction'), pin_direction
                            yield (
                                tile, site_name, bel_name, pin_name,
                                'is_clock'), int(pin_is_clock) == 1
                            yield (
                                tile, site_name, bel_name, pin_name,
                                'is_part_of_bus'
                            ), int(pin_is_part_of_bus) == 1
                            pin_loc += 4

    return merged_dict(inner())


def read_site_pins(pins_file):
    def inner():
        with open(pins_file, 'r') as f:
            for line in f:
                raw_pins = line.split()
                tile = raw_pins[0]
                site_count = int(raw_pins[1])
                pin_loc = 2

                if site_count == 0:
                    yield (tile, ), {}

                for site in range(0, site_count):
                    site_name = raw_pins[pin_loc]
                    site_name = site_name.lower()
                    site_pins_count = int(raw_pins[pin_loc + 1])

                    pin_loc += 2
                    for pin in range(0, site_pins_count):
                        pin_name = raw_pins[pin_loc]
                        pin_direction = raw_pins[pin_loc + 1]
                        pin_is_part_of_bus = raw_pins[pin_loc + 2]

                        yield (
                            (tile, site_name, pin_name, 'direction'),
                            pin_direction)
                        yield (
                            (tile, site_name, pin_name, 'is_clock'),
                            pin_name.lower() == 'clk')
                        yield (
                            (tile, site_name, pin_name, 'is_part_of_bus'),
                            int(pin_is_part_of_bus))

                        # site clock pins are always named 'CLK'
                        pin_loc += 3

    return merged_dict(inner())


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('--timings', type=str, help='Raw timing input file')
    parser.add_argument('--json', type=str, help='json output file')
    parser.add_argument(
        '--properties', type=str, help='Bel properties input file')
    parser.add_argument('--belpins', type=str, help='Bel pins input file')
    parser.add_argument('--sitepins', type=str, help='Site pins input file')
    parser.add_argument(
        '--debug', type=bool, default=False, help='Enable debug json dumps')
    parser.add_argument(
        '--propertiesmap', type=str, help='Properties names mappings')
    parser.add_argument(
        '--pinaliasmap', type=str, help='Pin name alias mappings')
    args = parser.parse_args()

    with open(args.propertiesmap, 'r') as fp:
        properties_map = json.load(fp)

    with open(args.pinaliasmap, 'r') as fp:
        pin_alias_map = json.load(fp)

    properties = read_bel_properties(args.properties, properties_map)

    if args.debug:
        with open('debug_prop.json', 'w') as fp:
            json.dump(properties, fp, indent=4, sort_keys=True)

    pins = read_bel_pins(args.belpins)
    if args.debug:
        with open('debug_pins.json', 'w') as fp:
            json.dump(pins, fp, indent=4, sort_keys=True)

    site_pins = read_site_pins(args.sitepins)
    if args.debug:
        with open('debug_site_pins.json', 'w') as fp:
            json.dump(site_pins, fp, indent=4, sort_keys=True)

    timings = read_raw_timings(
        args.timings, properties, pins, site_pins, pin_alias_map)
    with open(args.json, 'w') as fp:
        json.dump(timings, fp, indent=4, sort_keys=True)


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2020  The Project X-Ray Authors.
#
# Use of this source code is governed by a ISC-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/ISC
#
# SPDX-License-Identifier: ISC
import os
import itertools
import re
import random
random.seed(int(os.getenv("SEED"), 16))
from prjxray import util
from prjxray.db import Database

XY_RE = re.compile('^BUFHCE_X([0-9]+)Y([0-9]+)$')
BUFGCTRL_XY_RE = re.compile('^BUFGCTRL_X([0-9]+)Y([0-9]+)$')
"""
BUFHCE's can be driven from:

MMCME2_ADV
BUFHCE
PLLE2_ADV
BUFGCTRL
"""


def get_xy(s):
    m = BUFGCTRL_XY_RE.match(s)
    x = int(m.group(1))
    y = int(m.group(2))
    return x, y


def gen_sites(desired_site_type):
    db = Database(util.get_db_root(), util.get_part())
    grid = db.grid()

    for tile_name in sorted(grid.tiles()):
        loc = grid.loc_of_tilename(tile_name)
        gridinfo = grid.gridinfo_at_loc(loc)
        for site, site_type in gridinfo.sites.items():
            if site_type == desired_site_type:
                yield site


def gen_bufhce_sites():
    db = Database(util.get_db_root(), util.get_part())
    grid = db.grid()
    for tile_name in sorted(grid.tiles()):
        loc = grid.loc_of_tilename(tile_name)
        gridinfo = grid.gridinfo_at_loc(loc)
        sites = []

        for site, site_type in gridinfo.sites.items():
            if site_type == 'BUFHCE':
                sites.append(site)

        if sites:
            yield tile_name, set(sites)


def main():
    print('''
module top();
    ''')

    gclks = []
    for site in sorted(gen_sites("BUFGCTRL"), key=get_xy):
        wire_name = 'clk_{}'.format(site)
        gclks.append(wire_name)

        print(
            """
    wire {wire_name};
    (* KEEP, DONT_TOUCH, LOC = "{site}" *)
    BUFG bufg_{site} (
        .O({wire_name})
        );
        """.format(
                site=site,
                wire_name=wire_name,
            ))

    bufhce_sites = list(gen_bufhce_sites())

    opts = []
    for count in range(len(bufhce_sites)):
        for opt in itertools.combinations(bufhce_sites, count + 1):
            opts.append(opt)

    for gclk in gclks:
        for tile_name, sites in random.choice(opts):
            for site in sorted(sites):
                print(
                    """
        (* KEEP, DONT_TOUCH, LOC = "{site}" *)
        BUFHCE buf_{site} (
            .I({wire_name})
                );""".format(
                        site=site,
                        wire_name=gclk,
                    ))
                sites.remove(site)
                break

    print("endmodule")


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2020  The Project X-Ray Authors.
#
# Use of this source code is governed by a ISC-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/ISC
#
# SPDX-License-Identifier: ISC
import json
import os
import random
random.seed(int(os.getenv("SEED"), 16))
from prjxray import util
from prjxray import verilog
from prjxray.db import Database


def gen_sites():
    get_xy = util.create_xy_fun('BUFHCE_')
    db = Database(util.get_db_root(), util.get_part())
    grid = db.grid()
    for tile_name in sorted(grid.tiles()):
        loc = grid.loc_of_tilename(tile_name)
        gridinfo = grid.gridinfo_at_loc(loc)
        sites = []

        xs = []
        ys = []
        for site, site_type in gridinfo.sites.items():
            if site_type == 'BUFHCE':
                x, y = get_xy(site)
                xs.append(x)
                ys.append(y)

                sites.append((site, x, y))

        if sites:
            yield tile_name, min(xs), min(ys), sorted(sites)


def main():
    print('''
module top();
    ''')

    params_list = []
    for tile_name, x_min, y_min, sites in gen_sites():

        for site, x, y in sites:
            params = {}
            params['tile'] = tile_name
            params['site'] = site
            params['x'] = x - x_min
            params['y'] = y - y_min
            params['IN_USE'] = random.randint(0, 1)

            if params['IN_USE']:
                params['INIT_OUT'] = random.randint(0, 1)
                params['CE_TYPE'] = verilog.quote(
                    random.choice(('SYNC', 'ASYNC')))
                params['IS_CE_INVERTED'] = random.randint(0, 1)
                print(
                    '''
    (* KEEP, DONT_TOUCH, LOC = "{site}" *)
    BUFHCE #(
        .INIT_OUT({INIT_OUT}),
        .CE_TYPE({CE_TYPE}),
        .IS_CE_INVERTED({IS_CE_INVERTED})
        ) buf_{site} ();
                    '''.format(**params))

            params_list.append(params)

    print("endmodule")

    with open('params.json', 'w') as f:
        json.dump(params_list, f, indent=2)


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2020  The Project X-Ray Authors.
#
# Use of this source code is governed by a ISC-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/ISC
#
# SPDX-License-Identifier: ISC
import io
import re
import yaml
import json
import unittest

from utils import xjson


def load(f):
    data = f.read()
    if isinstance(data, bytes):
        data = data.decode('utf-8')
    # Strip out of !<tags>
    data = re.sub("!<[^>]*>", "", data)
    return yaml.safe_load(io.StringIO(data))


def tojson(f):
    d = load(f)
    o = io.StringIO()
    xjson.pprint(o, d)
    return o.getvalue()


class XYamlTest(unittest.TestCase):
    def test(self):
        s = io.StringIO(
            """\
!<xilinx/xc7series/part>
idcode: 0x362d093
global_clock_regions:
  top: !<xilinx/xc7series/global_clock_region>
    rows:
      0: !<xilinx/xc7series/row>
        configuration_buses:
          CLB_IO_CLK: !<xilinx/xc7series/configuration_bus>
            configuration_columns:
              0: !<xilinx/xc7series/configuration_column>
                frame_count: 42
""")
        djson = tojson(s)
        self.assertMultiLineEqual(
            djson, """\
{
    "global_clock_regions": {
        "top": {
            "rows": {
                "0": {
                    "configuration_buses": {
                        "CLB_IO_CLK": {
                            "configuration_columns": {
                                "0": {
                                    "frame_count": 42
                                }
                            }
                        }
                    }
                }
            }
        }
    },
    "idcode": 56807571
}""")


if __name__ == "__main__":
    import sys
    if len(sys.argv) == 1:
        unittest.main()
    else:
        assert len(sys.argv) == 2
        print(tojson(open(sys.argv[1])))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2020  The Project X-Ray Authors.
#
# Use of this source code is governed by a ISC-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/ISC
#
# SPDX-License-Identifier: ISC

import itertools
import os
import re

from prjxray import util

clb_int_zero_db = [
    # CLB interconnet
    # Ex:
    # segbits_hclk_l.db:73:HCLK_L.HCLK_LEAF_CLK_B_BOTL4.HCLK_CK_BUFHCLK10 00_21 04_21
    # segbits_int_l.db:207:INT_L.CLK_L0.GCLK_L_B8_WEST !01_21 00_21 00_25 01_20 01_24
    "00_21 00_22 00_26 01_28|00_25 01_20 01_21 01_24",
    "00_23 00_30 01_22 01_25|00_27 00_29 01_26 01_29",
    "01_12 01_14 01_16 01_18|00_10 00_11 01_09 01_10",
    "00_13 01_17 00_15 00_17|00_18 00_19 01_13 00_14",
    "00_34 00_38 01_33 01_37|00_35 00_39 01_38 01_40",
    "00_33 00_41 01_32 01_34|00_37 00_42 01_36 01_41",
]


def zero_range(tag, bits, wordmin, wordmax):
    """
    If any bits occur wordmin <= word <= wordmax,
    default bits in wordmin <= word <= wordmax to 0
    """

    # The bit index, if any, that needs to be one hotted
    bitidx = None
    for bit in bits:
        if bit[0] == "!":
            continue
        fidx, bidx = [int(s) for s in bit.split("_")]
        if wordmin <= fidx <= wordmax:
            if bitidx is not None and bidx != bitidx:
                print("Old bit index: %u, new: %u" % (bitidx, bidx))
                print("%s bits: %s" % (tag, str(bits)))
                raise ValueError("%s: inconsistent bit index" % tag)
            bitidx = bidx

    if bitidx is None:
        return

    for fidx in range(wordmin, wordmax + 1):
        bit = "%02d_%02d" % (fidx, bitidx)
        # Preserve 1 bits, set others to 0
        if bit not in bits:
            bits.add("!" + bit)


def bits_str(bits):
    """Convert a set into canonical form"""
    return ' '.join(sorted(list(bits)))


class ZeroGroups(object):
    def __init__(self, zero_db):
        self.groups = []
        self.bit_to_group = {}
        self.tag_to_groups = {}
        self.zero_tag_to_group = {}
        self.parse_zero_db(zero_db)

    def print_groups(self):
        print('Zero groups:')
        for bits in self.groups:
            print(bits_str(bits))

        print('Zero tags:')
        for tag in self.zero_tag_to_group:
            print(tag, bits_str(self.zero_tag_to_group[tag]))

    def parse_zero_db(self, zero_db):
        """ Convert zero db format into data structure

        Zero db format examples:

        Ex: 01_02 04_05
        Means find a line that has either of these bits
        If either of them occurs, default bits in that set to zero

        Ex: 01_02 04_05|07_08 10_11
        If any bits from the first group occur,
        default bits in the second group to zero

        Ex: 01_02 04_05,ALL_ZERO
        ALL_ZERO is an enum that is part of the group but is all 0
        It must have 0 candidates

        Ex: CLB.SLICE_X0.CLKINV ^ CLB.SLICE_X0.NOCLKINV
        CLB.SLICE_X0.NOCLKINV is all bits in CLB.SLICE_X0.CLKINV unset

        Ex: A | B ^ C
        C is all bits in (A)|(B) unset


        """
        for zdb in zero_db:

            if "^" in zdb:
                self.groups.append(set())
                zero_group = self.groups[-1]

                other_tags, allzero_tag = zdb.split('^')
                allzero_tag = allzero_tag.strip()

                for tag in other_tags.split():
                    self.tag_to_groups[tag.strip()] = [zero_group]

                self.zero_tag_to_group[allzero_tag] = zero_group
                continue

            allzero_tag = None
            if "," in zdb:
                zdb, allzero_tag = zdb.split(",")

            if "|" in zdb:
                a, b = zdb.split("|")
                a = a.split()
                b = b.split()

                self.groups.append(set(b))
                zero_group = self.groups[-1]
            else:
                a = zdb.split()
                self.groups.append(set(a))
                zero_group = self.groups[-1]

            if allzero_tag is not None:
                self.zero_tag_to_group[allzero_tag] = zero_group

            for bit in a:
                self.bit_to_group[bit] = zero_group

    def add_tag_bits(self, tag, bits):
        if tag in self.zero_tag_to_group:
            return

        group_ids = set()
        groups = []

        if tag in self.tag_to_groups:
            assert len(self.tag_to_groups[tag]) == 1

            self.tag_to_groups[tag][0] |= bits

            for bit in bits:
                if bit in self.bit_to_group:
                    # Make sure each bit only belongs to one group
                    assert id(self.bit_to_group[bit]) == id(
                        self.tag_to_groups[tag])
                else:
                    self.bit_to_group[bit] = self.tag_to_groups[tag]

            group_ids.add(id(self.tag_to_groups[tag]))
            groups = self.tag_to_groups[tag]

        for bit in bits:
            if bit in self.bit_to_group:
                if id(self.bit_to_group[bit]) not in group_ids:
                    group_ids.add(id(self.bit_to_group[bit]))
                    groups.append(self.bit_to_group[bit])

        self.tag_to_groups[tag] = groups

    def add_bits_from_zero_groups(self, tag, bits, strict=True, verbose=False):
        """ Add bits from a zero group, if needed

        Arguments
        ---------
        tag : str
            Tag being to examine for zero group
        bits : set of str
            Set of bits set on this tag
        strict : bool
            Assert that the size of the given group is the size of the given
            mask.
        verbose : bool
            Print to stdout grouping being made
        """

        tag_is_masked = tag in self.tag_to_groups
        tag_is_zero = tag in self.zero_tag_to_group

        # Should not have a tag that is both masked and a zero tag.
        assert not (tag_is_masked and tag_is_zero)

        if tag_is_masked:
            for b in self.tag_to_groups[tag]:
                bits_orig = set(bits)
                for bit in b:
                    if bit not in bits:
                        bits.add("!" + bit)

                verbose and print(
                    "Grouped %s: %s => %s" %
                    (tag, bits_str(bits_orig), bits_str(bits)))

        if tag_is_zero:
            for bit in self.zero_tag_to_group[tag]:
                bits.add("!" + bit)


def read_segbits(fn_in):
    """
    Reads a segbits file. Removes duplcated lines. Returns a list of the lines.
    """
    lines = []
    llast = None

    with util.OpenSafeFile(fn_in, "r") as f:
        for line in f:
            # Hack: skip duplicate lines
            # This happens while merging a new multibit entry
            line = line.strip()
            if len(line) == 0:
                continue
            if line == llast:
                continue

            lines.append(line)

    return lines


def add_zero_bits(
        fn_in, lines, zero_db, clb_int=False, strict=True, verbose=False):
    '''
    Add multibit entries
    This requires adding some zero bits (ex: !31_09)
    If an entry has any of the
    '''

    zero_groups = ZeroGroups(zero_db)

    new_lines = set()
    changes = 0

    drops = 0

    for line in lines:

        tag, bits, mode, _ = util.parse_db_line(line)

        if bits is not None and mode is None:
            zero_groups.add_tag_bits(tag, bits)

    if verbose:
        zero_groups.print_groups()

    for line in lines:
        tag, bits, mode, _ = util.parse_db_line(line)
        # an enum that needs masking
        # check below asserts that a mask was actually applied
        if mode and mode != "<0 candidates>" and not strict:
            verbose and print("WARNING: dropping unresolved line: %s" % line)
            drops += 1
            continue

        assert mode not in (
            "<const0>",
            "<const1>"), "Entries must be resolved. line: %s" % (line, )

        if mode == "always":
            new_line = line
        else:
            if mode:
                assert mode == "<0 candidates>", line
                bits = set()
            else:
                bits = set(bits)
            """
            This appears to be a large range of one hot interconnect bits
            They are immediately before the first CLB real bits
            """
            if clb_int:
                zero_range(tag, bits, 22, 25)

                set_bits = [bit for bit in bits if bit[0] != '!']
                if len(set_bits) not in [2, 4]:
                    # All INT bits appear to be only have 2 or 4 bits.
                    verbose and print(
                        "WARNING: dropping line with %d bits, not [2, 4]: %s, %s"
                        % (len(set_bits), bits, line))
                    drops += 1
                    continue

            zero_groups.add_bits_from_zero_groups(
                tag, bits, strict=strict, verbose=verbose)

            if strict:
                assert len(bits) > 0, 'Line {} found no bits.'.format(line)
            elif len(bits) == 0:
                verbose and print(
                    "WARNING: dropping unresolved line: %s" % line)
                drops += 1
                continue

            new_line = " ".join([tag] + sorted(bits))

        if re.match(r'.*<.*>.*', new_line):
            print("Original line: %s" % line)
            assert 0, "Failed to remove line mode: %s" % (new_line)

        if new_line != line:
            changes += 1
        new_lines.add(new_line)

    if drops:
        print("WARNING: %s dropped %s unresolved lines" % (fn_in, drops))

    return changes, new_lines


def update_mask(db_root, mask_db, src_dbs, offset=0):
    bits = set()
    mask_db_file = "%s/mask_%s.db" % (db_root, mask_db)

    if os.path.exists(mask_db_file):
        with util.OpenSafeFile(mask_db_file, "r") as f:
            for line in f:
                line = line.split()
                assert len(line) == 2
                assert line[0] == "bit"
                bits.add(line[1])

    for src_db in src_dbs:
        seg_db_file = "%s/segbits_%s.db" % (db_root, src_db)

        if not os.path.exists(seg_db_file):
            continue

        with util.OpenSafeFile(seg_db_file, "r") as f:
            for line in f:
                line = line.split()
                for bit in line[1:]:
                    if bit[0] == "!":
                        continue
                    if offset != 0:
                        m = re.match(r"(\d+)_(\d+)", bit)
                        bit = "%02d_%02d" % (
                            int(m.group(1)), int(m.group(2)) + offset)
                    bits.add(bit)

    if len(bits) > 0:
        with util.OpenSafeFile(mask_db_file, "w") as f:
            for bit in sorted(bits):
                print("bit %s" % bit, file=f)


def load_zero_db(fn):
    # Remove comments and convert to list of lines
    ret = []
    with util.OpenSafeFile(fn, "r") as f:
        for l in f:
            pos = l.find("#")
            if pos >= 0:
                l = l[0:pos]
            l = l.strip()
            if not l:
                continue
            ret.append(l)
    return ret


def remove_ambiguous_solutions(fn_in, db_lines, strict=True, verbose=True):
    """ Removes features with identical solutions.

    During solving, some tags may be tightly coupled and solve to the same
    solution.  In these cases, those solutions must be dropped until
    disambiguating information can be found.
    """
    solutions = {}
    dropped_solutions = set()

    for l in db_lines:
        parts = l.split()
        feature = parts[0]
        bits = frozenset(parts[1:])

        if bits in solutions:
            if strict:
                assert False, "Found solution {} at least twice, in {} and {}".format(
                    bits, feature, solutions[bits])
            else:
                dropped_solutions.add(bits)
        else:
            solutions[bits] = feature

    if strict:
        return 0, db_lines

    drops = 0
    output_lines = set()

    for l in db_lines:
        parts = l.split()
        feature = parts[0]
        bits = frozenset(parts[1:])

        if bits not in dropped_solutions:
            output_lines.add(l)
            drops += 1
        else:
            if verbose:
                print(
                    "WARNING: dropping line due to duplicate solution: %s" % l)

    if drops > 0:
        print("WARNING: %s dropped %s duplicate solutions" % (fn_in, drops))

    return drops, output_lines


def format_bits(tag, bits):
    """ Format tag and bits into line. """
    bit_strs = []
    for bit in sorted(list(bits), key=lambda b: b[1]):
        s = "!" if not bit[0] else ""
        s += "{:02d}_{:02d}".format(bit[1][0], bit[1][1])
        bit_strs.append(s)

    return " ".join([tag] + bit_strs)


def group_tags(lines, tag_groups, bit_groups):
    """
    Implements tag grouping. If a tag belongs to a group then the common bits
    of that group are added to is as zeros.

    >>> tg = [{"A", "B"}]
    >>> bg = [{(1, 2), (3, 4)}]
    >>> res = group_tags({"A 1_2", "B 3_4"}, tg, bg)
    >>> (res[0], sorted(list(res[1])))
    (2, ['A 01_02 !03_04', 'B !01_02 03_04'])

    >>> tg = [{"A", "B"}]
    >>> bg = [{(1, 2), (3, 4)}]
    >>> res = group_tags({"A 1_2", "B 3_4", "C 1_2"}, tg, bg)
    >>> (res[0], sorted(list(res[1])))
    (2, ['A 01_02 !03_04', 'B !01_02 03_04', 'C 01_02'])
    """

    changes = 0
    new_lines = set()

    # Process lines
    for line in lines:

        line = line.strip()
        if not len(line):
            continue

        # Parse the line
        tag, bits, mode, _ = util.parse_db_line(line)
        if not bits:
            bits = set()
        else:
            bits = set([util.parse_tagbit(b) for b in bits])

        # Check if the tag belongs to a group
        for tag_group, bit_group in zip(tag_groups, bit_groups):
            if tag in tag_group:

                # Add zero bits to the tag if not already there
                bit_coords = set([b[1] for b in bits])
                for zero_bit in bit_group:
                    if zero_bit not in bit_coords:
                        bits.add((False, zero_bit))

                # Format the line
                new_line = format_bits(tag, bits)

                # Add the line
                new_lines.add(new_line)
                changes += 1
                break

        # It does not, pass it through unchanged
        else:
            new_lines.add(format_bits(tag, bits))

    return changes, new_lines


def update_seg_fns(
        fn_inouts,
        zero_db,
        tag_groups,
        clb_int,
        lazy=False,
        strict=True,
        verbose=False):

    seg_files = 0
    seg_lines = 0
    for fn_in, fn_out in fn_inouts:
        verbose and print("zb %s: %s" % (fn_in, os.path.exists(fn_in)))
        if lazy and not os.path.exists(fn_in):
            continue

        lines = read_segbits(fn_in)
        changes = 0

        # Find common bits for tag groups
        bit_groups = find_common_bits_for_tag_groups(lines, tag_groups)

        # Group tags
        new_changes, lines = group_tags(lines, tag_groups, bit_groups)
        changes += new_changes

        new_changes, lines = add_zero_bits(
            fn_in,
            lines,
            zero_db,
            clb_int=clb_int,
            strict=strict,
            verbose=verbose)
        changes += new_changes

        new_changes, lines = remove_ambiguous_solutions(
            fn_in,
            lines,
            strict=strict,
            verbose=verbose,
        )
        changes += new_changes

        with util.OpenSafeFile(fn_out, "w") as f:
            for line in sorted(lines):
                print(line, file=f)

        if changes is not None:
            seg_files += 1
            seg_lines += changes
    print(
        "Segbit: checked %u files w/ %u changed lines" %
        (seg_files, seg_lines))


def update_masks(db_root):
    for mask_db, src_dbs in [
        ("clbll_l", ("clbll_l", "int_l")),
        ("clbll_r", ("clbll_r", "int_r")),
        ("clblm_l", ("clblm_l", "int_l")),
        ("clblm_r", ("clblm_r", "int_r")),
        ("hclk_l", ("hclk_l", )),
        ("hclk_r", ("hclk_r", )),
        ("bram_l", ("bram_l", )),
        ("bram_r", ("bram_r", )),
        ("dsp_l", ("dsp_l", )),
        ("dsp_r", ("dsp_r", )),
    ]:
        update_mask(db_root, mask_db, src_dbs)

    for mask_db, src_dbs in [
        ("bram_l", ("int_l", )),
        ("bram_r", ("int_r", )),
        ("dsp_l", ("int_l", )),
        ("dsp_r", ("int_r", )),
    ]:
        for k in range(5):
            update_mask(db_root, mask_db, src_dbs, offset=64 * k)

    print("Mask: checked files")


def update_segs(
        db_root,
        clb_int,
        seg_fn_in,
        seg_fn_out,
        zero_db_fn,
        tag_groups,
        strict=True,
        verbose=False):
    if clb_int:
        zero_db = clb_int_zero_db
        lazy = True

        def gen_fns():
            for tile_type in ["int_l", "int_r", "clbll_l", "clbll_r",
                              "clblm_l", "clblm_r"]:
                fn = "%s/segbits_%s.db" % (db_root, tile_type)
                yield (fn, fn)

        fn_inouts = list(gen_fns())
    else:
        assert seg_fn_in
        assert zero_db_fn
        lazy = False

        if not seg_fn_out:
            seg_fn_out = seg_fn_in

        fn_inouts = [(seg_fn_in, seg_fn_out)]
        zero_db = load_zero_db(zero_db_fn)
    print("CLB INT mode: %s" % clb_int)
    print("Segbit groups: %s" % len(zero_db))
    update_seg_fns(
        fn_inouts,
        zero_db,
        tag_groups,
        clb_int,
        lazy=lazy,
        strict=strict,
        verbose=verbose)


def find_common_bits_for_tag_groups(lines, tag_groups):
    """
    For each tag group finds a common set of bits that have value of one.
    """

    bit_groups = []

    for tag_group in tag_groups:
        bit_group = set()

        for line in lines:
            tag, bits, mode, _ = util.parse_db_line(line)
            if not bits:
                continue

            bits = set([util.parse_tagbit(b) for b in bits])

            if tag in tag_group and len(bits):
                ones = set([b[1] for b in bits if b[0]])
                bit_group |= ones

        bit_groups.append(bit_group)

    return bit_groups


def load_tag_groups(file_name):
    """
    Loads tag groups from a text file.

    A tag group is defined by specifying a space separated list of tags within
    a single line. Lines that are empty or start with '#' are ignored.
    """
    tag_groups = []

    # Load tag group specifications
    with util.OpenSafeFile(file_name, "r") as fp:
        for line in fp:
            line = line.strip()

            if len(line) == 0 or line.startswith("#"):
                continue

            group = set(line.split())
            if len(group):
                tag_groups.append(group)

    # Check if all tag groups are exclusive
    for tag_group_a, tag_group_b in itertools.combinations(tag_groups, 2):

        tags = tag_group_a & tag_group_b
        if len(tags):
            raise RuntimeError(
                "Tag(s) {} are present in multiple groups".format(
                    " ".join(tags)))

    return tag_groups


def run(
        db_root,
        clb_int=False,
        zero_db_fn=None,
        seg_fn_in=None,
        seg_fn_out=None,
        groups_fn_in=None,
        strict=None,
        verbose=False):

    if strict is None:
        strict = not clb_int

    # Load tag groups
    tag_groups = []
    if groups_fn_in is not None:
        tag_groups = load_tag_groups(groups_fn_in)

    # Probably should split this into two programs
    update_segs(
        db_root,
        clb_int=clb_int,
        seg_fn_in=seg_fn_in,
        seg_fn_out=seg_fn_out,
        zero_db_fn=zero_db_fn,
        tag_groups=tag_groups,
        strict=strict,
        verbose=verbose)
    if clb_int:
        update_masks(db_root)


def main():
    import argparse

    parser = argparse.ArgumentParser(description='Create multi-bit entries')

    util.db_root_arg(parser)
    parser.add_argument('--verbose', action='store_true', help='')
    parser.add_argument(
        '--clb-int', action='store_true', help='Fixup CLB interconnect')
    parser.add_argument('--zero-db', help='Apply custom patches')
    parser.add_argument('--seg-fn-in', help='')
    parser.add_argument('--seg-fn-out', help='')
    util.add_bool_arg(parser, "--strict", default=False)

    parser.add_argument(
        "-g",
        "--groups",
        type=str,
        default=None,
        help="Input tag group definition file")

    args = parser.parse_args()

    run(
        args.db_root,
        args.clb_int,
        args.zero_db,
        args.seg_fn_in,
        args.seg_fn_out,
        args.groups,
        strict=args.strict,
        verbose=args.verbose)


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright (C) 2017-2020  The Project X-Ray Authors.
#
# Use of this source code is governed by a ISC-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/ISC
#
# SPDX-License-Identifier: ISC

import os
import json
import random
random.seed(int(os.getenv("SEED"), 16))
from prjxray import util


def gen_sites():
    for tile_name, site_name, _site_type in sorted(util.get_roi().gen_sites(
        ['RAMB18E1', 'FIFO18E1'])):
        yield tile_name, site_name


BITS_PER_PARAM = 256
NUM_INITP_PARAMS = 8
NUM_INIT_PARAMS = 0x40
BITS_PER_SITE = BITS_PER_PARAM * (NUM_INITP_PARAMS + NUM_INIT_PARAMS)


def main():
    print("module top();")

    list_of_params = []
    for tile_name, site in gen_sites():
        params = {}
        params['tile'] = tile_name
        params['site'] = site

        p = []
        for initp_idx in range(NUM_INITP_PARAMS):
            param = 'INITP_{:02X}'.format(initp_idx)
            params[param] = random.randint(0, 2**BITS_PER_PARAM - 1)
            p.append(
                ".{param}(256'h{val:x})".format(
                    param=param, val=params[param]))

        for init_idx in range(NUM_INIT_PARAMS):
            param = 'INIT_{:02X}'.format(init_idx)
            params[param] = random.randint(0, 2**BITS_PER_PARAM - 1)
            p.append(
                ".{param}(256'h{val:x})".format(
                    param=param, val=params[param]))

        params['params'] = ','.join(p)

        print(
            """
    (* KEEP, DONT_TOUCH, LOC = "{site}" *)
    RAMB18E1 #(
        .READ_WIDTH_A(1),
        .READ_WIDTH_B(1),
        .RAM_MODE("TDP"),
        {params}
        ) bram_{site} (
        );
        """.format(**params))

        list_of_params.append(params)

    print("endmodule")

    with open('params.json', 'w') as f:
        json.dump(list_of_params, f, indent=2)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2020  The Project X-Ray Authors.
#
# Use of this source code is governed by a ISC-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/ISC
#
# SPDX-License-Identifier: ISC

import sys, re

from prjxray.segmaker import Segmaker

pipdata = dict()
ignpip = set()


def handle_design(prefix, second_pass):
    segmk = Segmaker(prefix + ".bits")

    tiledata = dict()
    nlines = 0

    print("Loading tags from design.txt.")
    with open(prefix + ".txt", "r") as f:
        for line in f:
            tile, pip, src, dst, pnum, pdir = line.split()
            _, pip = pip.split(".")
            _, src = src.split("/")
            _, dst = dst.split("/")
            pnum = int(pnum)
            pdir = int(pdir)

            if tile not in tiledata:
                tiledata[tile] = {"pips": set(), "srcs": set(), "dsts": set()}

            if pip in pipdata:
                assert pipdata[pip] == (src, dst)
            else:
                pipdata[pip] = (src, dst)

            tiledata[tile]["pips"].add(pip)
            tiledata[tile]["srcs"].add(src)
            tiledata[tile]["dsts"].add(dst)

            if pdir == 0:
                tiledata[tile]["srcs"].add(dst)
                tiledata[tile]["dsts"].add(src)

            if pnum == 1 or pdir == 0 or \
                    re.match(r"^(L[HV]B?)(_L)?(_B)?[0-9]", src) or \
                    re.match(r"^(L[HV]B?)(_L)?(_B)?[0-9]", dst) or \
                    re.match(r"^(CTRL|GFAN)(_L)?[0-9]", dst):
                ignpip.add(pip)

            nlines += 1

    if nlines == 0:
        return

    for tile, pips_srcs_dsts in tiledata.items():
        pips = pips_srcs_dsts["pips"]
        srcs = pips_srcs_dsts["srcs"]
        dsts = pips_srcs_dsts["dsts"]

        for pip, src_dst in pipdata.items():
            src, dst = src_dst
            if pip in ignpip:
                pass
            elif pip in pips:
                segmk.addtag(tile, "%s.%s" % (dst, src), 1)
            elif src_dst[1] not in dsts:
                segmk.addtag(tile, "%s.%s" % (dst, src), 0)

    if second_pass:
        segmk.compile()
        segmk.write(prefix[7:])


for arg in sys.argv[1:]:
    prefix = arg[0:-4]
    handle_design(prefix, False)

for arg in sys.argv[1:]:
    prefix = arg[0:-4]
    handle_design(prefix, True)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2020  The Project X-Ray Authors.
#
# Use of this source code is governed by a ISC-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/ISC
#
# SPDX-License-Identifier: ISC
import json
import io
import os
import random
random.seed(int(os.getenv("SEED"), 16))
from prjxray import util
from prjxray import lut_maker
from prjxray import verilog
from prjxray.db import Database


def gen_sites():
    '''
    IOB33S: main IOB of a diff pair
    IOB33M: secondary IOB of a diff pair
    IOB33: not a diff pair. Relatively rare (at least in ROI...2 of them?)
    Focus on IOB33S to start
    '''
    db = Database(util.get_db_root(), util.get_part())
    grid = db.grid()
    for tile_name in sorted(grid.tiles()):
        loc = grid.loc_of_tilename(tile_name)
        gridinfo = grid.gridinfo_at_loc(loc)

        for site_name, site_type in gridinfo.sites.items():
            if site_type in ['IOB33S', 'IOB33M']:
                yield tile_name, site_name


def write_params(params):
    pinstr = 'tile,site,pin,iostandard,drive,slew\n'
    for vals in params:
        pinstr += ','.join(map(str, vals)) + '\n'

    open('params.csv', 'w').write(pinstr)


def use_iserdese2(p, luts, connects):
    iobdelay = random.choice((
        'NONE',
        'BOTH',
        'IBUF',
        'IFD',
    ))

    p['IOBDELAY'] = verilog.quote(iobdelay)
    p['INIT_Q1'] = random.randint(0, 1)
    p['INIT_Q2'] = random.randint(0, 1)
    p['INIT_Q3'] = random.randint(0, 1)
    p['INIT_Q4'] = random.randint(0, 1)

    p['SRVAL_Q1'] = random.randint(0, 1)
    p['SRVAL_Q2'] = random.randint(0, 1)
    p['SRVAL_Q3'] = random.randint(0, 1)
    p['SRVAL_Q4'] = random.randint(0, 1)
    p['NUM_CE'] = random.randint(1, 2)

    p['IS_CLK_INVERTED'] = random.randint(0, 1)
    p['IS_CLKB_INVERTED'] = random.randint(0, 1)
    p['IS_OCLK_INVERTED'] = random.randint(0, 1)
    p['IS_OCLKB_INVERTED'] = random.randint(0, 1)
    p['IS_CLKDIV_INVERTED'] = random.randint(0, 1)
    p['IS_D_INVERTED'] = random.randint(0, 1)
    p['INTERFACE_TYPE'] = verilog.quote(
        random.choice(
            (
                'MEMORY',
                'MEMORY_DDR3',
                'MEMORY_QDR',
                'NETWORKING',
                'OVERSAMPLE',
            )))
    p['DATA_RATE'] = verilog.quote(random.choice((
        'SDR',
        'DDR',
    )))
    if verilog.unquote(p['DATA_RATE']) == 'SDR':
        data_widths = [2, 3, 4, 5, 6, 7, 8]
    else:
        data_widths = [4, 6, 8]

    p['DATA_WIDTH'] = random.choice(data_widths)
    p['SERDES_MODE'] = verilog.quote(random.choice(('MASTER', 'SLAVE')))

    use_delay = iobdelay != 'NONE'

    if iobdelay == 'NONE':
        p['mux_config'] = 'direct'
        p['iddr_mux_config'] = 'direct'
    elif iobdelay == 'BOTH':
        p['mux_config'] = 'idelay'
        p['iddr_mux_config'] = 'idelay'
    elif iobdelay == 'IBUF':
        p['mux_config'] = 'idelay'
        p['iddr_mux_config'] = 'direct'
    elif iobdelay == 'IFD':
        p['mux_config'] = 'direct'
        p['iddr_mux_config'] = 'idelay'

    p['OFB_USED'] = verilog.quote(random.choice(('TRUE', 'FALSE')))
    p['DYN_CLKDIV_INV_EN'] = verilog.quote(random.choice(('TRUE', 'FALSE')))
    p['DYN_CLK_INV_EN'] = verilog.quote(random.choice(('TRUE', 'FALSE')))

    if use_delay:
        print(
            """
    wire idelay_{site};

    (* KEEP, DONT_TOUCH, LOC = "{idelay_loc}" *)
    IDELAYE2 #(
    ) idelay_site_{site} (
        .IDATAIN({iwire}),
        .DATAOUT(idelay_{site})
        );""".format(**p),
            file=connects)

        p['ddly_connection'] = '.DDLY(idelay_{site}),'.format(**p)
    else:
        p['ddly_connection'] = ''

    if verilog.unquote(p['OFB_USED']) == 'TRUE':
        p['ODATA_RATE'] = verilog.quote(random.choice((
            'SDR',
            'DDR',
        )))
        if verilog.unquote(p['ODATA_RATE']) == 'SDR':
            data_widths = [2, 3, 4, 5, 6, 7, 8]
        else:
            data_widths = [4, 6, 8]

        p['ODATA_WIDTH'] = random.choice(data_widths)
        p['OSERDES_MODE'] = verilog.quote(random.choice(('MASTER', 'SLAVE')))

        if p['ODATA_WIDTH'] == 4 and verilog.unquote(p['ODATA_RATE']) == 'DDR':
            p['TRISTATE_WIDTH'] = 4
        else:
            p['TRISTATE_WIDTH'] = 1

        print(
            """
    wire tfb_{site};
    wire ofb_{site};

    (* KEEP, DONT_TOUCH, LOC = "{ologic_loc}" *)
    OSERDESE2 #(
        .SERDES_MODE({OSERDES_MODE}),
        .DATA_RATE_TQ({ODATA_RATE}),
        .DATA_RATE_OQ({ODATA_RATE}),
        .DATA_WIDTH({ODATA_WIDTH}),
        .TRISTATE_WIDTH({TRISTATE_WIDTH})
    ) oserdese2_{site} (
        .CLK(0),
        .CLKDIV(0),
        .D1(0),
        .TFB(tfb_{site}),
        .OQ({owire}),
        .TQ({twire}),
        .OFB(ofb_{site})
        );""".format(**p),
            file=connects)

        p['ofb_connections'] = """
        .OFB(ofb_{site}),
        """.format(**p)
    else:
        p['ofb_connections'] = ''

    if random.randint(0, 1):
        clknet = luts.get_next_output_net()
    else:
        clknet = random.choice((
            'clk_BUFG1',
            'clk_BUFG2',
        ))

    if random.randint(0, 1):
        clkbnet = luts.get_next_output_net()
    else:
        clkbnet = random.choice((
            'clk_BUFG1',
            'clk_BUFG2',
        ))

    if random.randint(0, 1):
        oclknet = luts.get_next_output_net()
    else:
        oclknet = random.choice((
            'clk_BUFG1',
            'clk_BUFG2',
        ))

    clkdiv = random.choice(('clk_BUFG3', 'clk_BUFG4'))

    p['DISABLE_CLOCKS'] = random.randint(0, 1)
    if p['DISABLE_CLOCKS']:
        clknet = '0'
        clkbnet = '0'
        oclknet = '0'
        clkdiv = '0'

    print(
        '''
    (* KEEP, DONT_TOUCH, LOC = "{ilogic_loc}" *)
    ISERDESE2 #(
        .SERDES_MODE({SERDES_MODE}),
        .INIT_Q1({INIT_Q1}),
        .INIT_Q2({INIT_Q2}),
        .INIT_Q3({INIT_Q3}),
        .INIT_Q4({INIT_Q4}),
        .SRVAL_Q1({SRVAL_Q1}),
        .SRVAL_Q2({SRVAL_Q2}),
        .SRVAL_Q3({SRVAL_Q3}),
        .SRVAL_Q4({SRVAL_Q4}),
        .DYN_CLKDIV_INV_EN({DYN_CLKDIV_INV_EN}),
        .DYN_CLK_INV_EN({DYN_CLK_INV_EN}),
        .INTERFACE_TYPE({INTERFACE_TYPE}),
        .IS_CLK_INVERTED({IS_CLK_INVERTED}),
        .IS_CLKB_INVERTED({IS_CLKB_INVERTED}),
        .IS_OCLK_INVERTED({IS_OCLK_INVERTED}),
        .IS_OCLKB_INVERTED({IS_OCLKB_INVERTED}),
        .IS_CLKDIV_INVERTED({IS_CLKDIV_INVERTED}),
        .IS_D_INVERTED({IS_D_INVERTED}),
        .OFB_USED({OFB_USED}),
        .NUM_CE({NUM_CE}),
        .DATA_RATE({DATA_RATE}),
        .DATA_WIDTH({DATA_WIDTH}),
        .IOBDELAY({IOBDELAY})
    ) iserdese2_{site} (
        {ddly_connection}
        {ofb_connections}
        .D({iwire}),
        .CLK({clknet}),
        .CLKB({clkbnet}),
        .OCLK({oclknet}),
        .O({onet}),
        .Q1({q1net}),
        .CLKDIV({clkdiv})
    );'''.format(
            clkdiv=clkdiv,
            clknet=clknet,
            clkbnet=clkbnet,
            oclknet=oclknet,
            onet=luts.get_next_input_net(),
            q1net=luts.get_next_input_net(),
            shiftout1net=luts.get_next_input_net(),
            shiftout2net=luts.get_next_input_net(),
            **p),
        file=connects)


def use_direct_and_iddr(p, luts, connects):
    p['mux_config'] = random.choice((
        'direct',
        'idelay',
        'none',
    ))

    p['iddr_mux_config'] = random.choice((
        'direct',
        'idelay',
        'none',
    ))

    if p['iddr_mux_config'] != 'none':
        p['INIT_Q1'] = random.randint(0, 1)
        p['INIT_Q2'] = random.randint(0, 1)
        p['IS_C_INVERTED'] = random.randint(0, 1)
        p['IS_D_INVERTED'] = random.randint(0, 1)
        p['SRTYPE'] = verilog.quote(random.choice(('SYNC', 'ASYNC')))
        p['DDR_CLK_EDGE'] = verilog.quote(
            random.choice(
                (
                    'OPPOSITE_EDGE',
                    'SAME_EDGE',
                    'SAME_EDGE_PIPELINED',
                )))

        print(
            '''
    (* KEEP, DONT_TOUCH, LOC = "{ilogic_loc}" *)
    IDDR #(
        .IS_D_INVERTED({IS_D_INVERTED}),
        .IS_C_INVERTED({IS_C_INVERTED}),
        .INIT_Q1({INIT_Q1}),
        .INIT_Q2({INIT_Q2}),
        .SRTYPE({SRTYPE}),
        .DDR_CLK_EDGE({DDR_CLK_EDGE})
    ) iddr_{site} (
        .C({cnet}),
        .D(iddr_d_{site}),
        .Q1({q1}),
        .Q2({q2})
        );
        '''.format(
                cnet=luts.get_next_output_net(),
                q1=luts.get_next_input_net(),
                q2=luts.get_next_input_net(),
                **p),
            file=connects)

    if p['iddr_mux_config'] == 'idelay' or p['mux_config'] == 'idelay' or p[
            'iddr_mux_config'] == 'tristate_feedback':
        print(
            """
    wire idelay_{site};

    (* KEEP, DONT_TOUCH, LOC = "{idelay_loc}" *)
    IDELAYE2 #(
    ) idelay_site_{site} (
        .IDATAIN({iwire}),
        .DATAOUT(idelay_{site})
        );""".format(**p),
            file=connects)

    print(
        """
    assign {owire} = {onet};
    assign {twire} = {tnet};
        """.format(
            onet=luts.get_next_output_net(),
            tnet=luts.get_next_output_net(),
            **p),
        file=connects)

    if p['iddr_mux_config'] == 'direct':
        print(
            '''
    assign iddr_d_{site} = {iwire};'''.format(**p, ),
            file=connects)
    elif p['iddr_mux_config'] == 'idelay':
        print(
            '''
    assign iddr_d_{site} = idelay_{site};'''.format(**p, ),
            file=connects)
    elif p['iddr_mux_config'] == 'tristate_feedback':
        print(
            '''
    assign iddr_d_{site} = tfb_{site} ? ofb_{site} : idelay_{site};'''.format(
                **p, ),
            file=connects)
    elif p['iddr_mux_config'] == 'none':
        pass
    else:
        assert False, p['mux_config']

    if p['mux_config'] == 'direct':
        print(
            '''
    assign {net} = {iwire};'''.format(
                net=luts.get_next_input_net(),
                **p,
            ),
            file=connects)
    elif p['mux_config'] == 'idelay':
        print(
            '''
    assign {net} = idelay_{site};'''.format(
                net=luts.get_next_input_net(),
                **p,
            ),
            file=connects)
    elif p['mux_config'] == 'none':
        pass
    else:
        assert False, p['mux_config']


def run():
    iostandards = [
        'LVCMOS12', 'LVCMOS15', 'LVCMOS18', 'LVCMOS25', 'LVCMOS33', 'LVTTL'
    ]
    iostandard = random.choice(iostandards)

    if iostandard in ['LVTTL', 'LVCMOS18']:
        drives = [4, 8, 12, 16, 24]
    elif iostandard == 'LVCMOS12':
        drives = [4, 8, 12]
    else:
        drives = [4, 8, 12, 16]

    slews = ['FAST', 'SLOW']
    pulls = ["NONE", "KEEPER", "PULLDOWN", "PULLUP"]

    luts = lut_maker.LutMaker()

    connects = io.StringIO()

    tile_params = []
    params = []
    for idx, (tile, site) in enumerate(gen_sites()):
        if idx == 0:
            continue

        p = {}
        p['tile'] = tile
        p['site'] = site
        p['ilogic_loc'] = site.replace('IOB', 'ILOGIC')
        p['ologic_loc'] = site.replace('IOB', 'OLOGIC')
        p['idelay_loc'] = site.replace('IOB', 'IDELAY')
        p['IOSTANDARD'] = verilog.quote(iostandard)
        p['PULLTYPE'] = verilog.quote(random.choice(pulls))
        p['DRIVE'] = random.choice(drives)
        p['SLEW'] = verilog.quote(random.choice(slews))

        p['pad_wire'] = 'dio[{}]'.format(idx - 1)
        p['owire'] = 'do_buf[{}]'.format(idx - 1)
        p['iwire'] = 'di_buf[{}]'.format(idx - 1)
        p['twire'] = 't[{}]'.format(idx - 1)

        params.append(p)
        tile_params.append(
            (
                tile, site, p['pad_wire'], iostandard, p['DRIVE'],
                verilog.unquote(p['SLEW']) if p['SLEW'] else None,
                verilog.unquote(p['PULLTYPE'])))

    write_params(tile_params)

    print(
        '''
`define N_DI {n_di}

module top(input clk, inout wire [`N_DI-1:0] dio);
    wire [`N_DI-1:0] di_buf;
    wire [`N_DI-1:0] do_buf;
    wire [`N_DI-1:0] t;

    wire clk_BUFG1;
    wire clk_BUFG2;
    wire clk_BUFG3;
    wire clk_BUFG4;

    (* KEEP, DONT_TOUCH  *)
    BUFG bufg1(
        .O(clk_BUFG1)
        );
    (* KEEP, DONT_TOUCH  *)
    BUFG bufg2(
        .O(clk_BUFG2)
        );
    (* KEEP, DONT_TOUCH  *)
    BUFG bufg3(
        .O(clk_BUFG3)
        );
    (* KEEP, DONT_TOUCH  *)
    BUFG bufg4(
        .O(clk_BUFG4)
        );
    '''.format(n_di=idx))

    # Always output a LUT6 to make placer happy.
    print(
        '''
        (* KEEP, DONT_TOUCH *)
        LUT6 dummy_lut();
        ''')

    any_idelay = False

    for p in params:
        print(
            '''
        wire iddr_d_{site};

        (* KEEP, DONT_TOUCH, LOC = "{site}" *)
        IOBUF #(
            .IOSTANDARD({IOSTANDARD})
        ) ibuf_{site} (
            .IO({pad_wire}),
            .I({owire}),
            .O({iwire}),
            .T({twire})
            );
            '''.format(**p),
            file=connects)

        p['use_iserdese2'] = random.randint(0, 1)
        if p['use_iserdese2']:
            use_iserdese2(p, luts, connects)
        else:
            use_direct_and_iddr(p, luts, connects)

        if p['iddr_mux_config'] == 'idelay' or p['mux_config'] == 'idelay':
            any_idelay = True

    if any_idelay:
        print("""
    (* KEEP, DONT_TOUCH *)
    IDELAYCTRL();""")

    for l in luts.create_wires_and_luts():
        print(l)

    print(connects.getvalue())

    print("endmodule")

    with open('params.jl', 'w') as f:
        json.dump(params, f, indent=2)


if __name__ == '__main__':
    run()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2020  The Project X-Ray Authors.
#
# Use of this source code is governed by a ISC-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/ISC
#
# SPDX-License-Identifier: ISC
import json

from prjxray.segmaker import Segmaker, add_site_group_zero
from prjxray import verilog


def bitfilter(frame, word):
    if frame < 26:
        return False

    return True


def main():
    segmk = Segmaker("design.bits", verbose=True)

    # Load tags
    with open("params.json", "r") as fp:
        data = json.load(fp)

    idelay_types = ["FIXED", "VARIABLE", "VAR_LOAD"]
    delay_srcs = ["IDATAIN", "DATAIN"]

    # Output tags
    for params in data:
        segmk.add_site_tag(params['IDELAY_IN_USE'], 'IN_USE', True)
        segmk.add_site_tag(params['IDELAY_NOT_IN_USE'], 'IN_USE', False)

        loc = verilog.unquote(params["LOC"])

        # Delay type
        value = verilog.unquote(params["IDELAY_TYPE"])
        value = value.replace(
            "_PIPE", "")  # VAR_LOAD and VAR_LOAD_PIPE are the same
        add_site_group_zero(
            segmk, loc, "IDELAY_TYPE_", idelay_types, "FIXED", value)

        # Delay value
        value = int(params["IDELAY_VALUE"])
        for i in range(5):
            segmk.add_site_tag(
                loc, "IDELAY_VALUE[%01d]" % i, ((value >> i) & 1) != 0)
            segmk.add_site_tag(
                loc, "ZIDELAY_VALUE[%01d]" % i, ((value >> i) & 1) == 0)

        # Delay source
        value = verilog.unquote(params["DELAY_SRC"])
        for x in delay_srcs:
            segmk.add_site_tag(loc, "DELAY_SRC_%s" % x, int(value == x))

        value = verilog.unquote(params["CINVCTRL_SEL"])
        segmk.add_site_tag(loc, "CINVCTRL_SEL", int(value == "TRUE"))

        value = verilog.unquote(params["PIPE_SEL"])
        segmk.add_site_tag(loc, "PIPE_SEL", int(value == "TRUE"))

        if "IS_C_INVERTED" in params:
            segmk.add_site_tag(
                loc, "IS_C_INVERTED", int(params["IS_C_INVERTED"]))
            segmk.add_site_tag(loc, "ZINV_C", 1 ^ int(params["IS_C_INVERTED"]))

        segmk.add_site_tag(
            loc, "IS_DATAIN_INVERTED", int(params["IS_DATAIN_INVERTED"]))
        if params['IBUF_IN_USE']:
            value = verilog.unquote(params["HIGH_PERFORMANCE_MODE"])
            segmk.add_site_tag(
                loc, "HIGH_PERFORMANCE_MODE", int(value == "TRUE"))

            segmk.add_site_tag(
                loc, "IS_IDATAIN_INVERTED", int(params["IS_IDATAIN_INVERTED"]))

    segmk.compile(bitfilter=bitfilter)
    segmk.write()


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2020  The Project X-Ray Authors.
#
# Use of this source code is governed by a ISC-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/ISC
#
# SPDX-License-Identifier: ISC

import json
import os
import random
from collections import namedtuple

random.seed(int(os.getenv("SEED"), 16))
from prjxray import util
from prjxray import verilog
from prjxray.lut_maker import LutMaker
from prjxray.db import Database

INT = "INT"
BIN = "BIN"
BOOL = "BOOL"
STR = "STR"


def gen_sites(site):
    db = Database(util.get_db_root(), util.get_part())
    grid = db.grid()
    already_used = list()
    for tile_name in sorted(grid.tiles()):
        loc = grid.loc_of_tilename(tile_name)
        gridinfo = grid.gridinfo_at_loc(loc)

        if gridinfo.tile_type not in [
                "GTP_CHANNEL_0",
                "GTP_CHANNEL_1",
                "GTP_CHANNEL_2",
                "GTP_CHANNEL_3",
                "GTP_CHANNEL_0_MID_LEFT",
                "GTP_CHANNEL_1_MID_LEFT",
                "GTP_CHANNEL_2_MID_LEFT",
                "GTP_CHANNEL_3_MID_LEFT",
                "GTP_CHANNEL_0_MID_RIGHT",
                "GTP_CHANNEL_1_MID_RIGHT",
                "GTP_CHANNEL_2_MID_RIGHT",
                "GTP_CHANNEL_3_MID_RIGHT",
        ] or gridinfo.tile_type in already_used:
            continue
        else:
            tile_type = gridinfo.tile_type
            already_used.append(tile_type)

        for site_name, site_type in gridinfo.sites.items():
            if site_type != site:
                continue

            if "RIGHT" in tile_type and "X0" in site_name:
                continue

            if "LEFT" in tile_type and "X1" in site_name:
                continue

            yield tile_name, tile_type, site_name, site_type


def main():
    print(
        '''
module top(
    input wire in,
    output wire out
);

assign out = in;
''')

    luts = LutMaker()

    primitives_list = list()

    for tile_name, tile_type, site_name, site_type in gen_sites(
            "GTPE2_CHANNEL"):

        params_list = list()
        params_dict = dict()

        params_dict["tile_type"] = tile_type
        params = dict()
        params['site'] = site_name

        verilog_attr = ""

        verilog_attr = "#("

        fuz_dir = os.getenv("FUZDIR", None)
        assert fuz_dir
        with open(os.path.join(fuz_dir, "attrs.json"), "r") as attrs_file:
            attrs = json.load(attrs_file)

        in_use = bool(random.randint(0, 9))
        params["IN_USE"] = in_use

        if in_use:
            for param, param_info in attrs.items():
                param_type = param_info["type"]
                param_values = param_info["values"]
                param_digits = param_info["digits"]

                if param_type == INT:
                    value = random.choice(param_values)
                    value_str = value
                elif param_type == BIN:
                    value = random.randint(0, param_values[0])
                    value_str = "{digits}'b{value:0{digits}b}".format(
                        value=value, digits=param_digits)
                elif param_type in [BOOL, STR]:
                    value = random.choice(param_values)
                    value_str = verilog.quote(value)

                params[param] = value

                verilog_attr += """
            .{}({}),""".format(param, value_str)

            verilog_ports = ""
            for param in ["TXUSRCLK", "TXUSRCLK2", "TXPHDLYTSTCLK",
                          "SIGVALIDCLK", "RXUSRCLK", "RXUSRCLK2", "DRPCLK",
                          "DMONITORCLK", "CLKRSVD0", "CLKRSVD1"]:
                is_inverted = random.randint(0, 1)

                params[param] = is_inverted

                verilog_attr += """
            .IS_{}_INVERTED({}),""".format(param, is_inverted)
                verilog_ports += """
            .{}({}),""".format(param, luts.get_next_output_net())

            verilog_attr = verilog_attr.rstrip(",")
            verilog_attr += "\n)"

            print("(* KEEP, DONT_TOUCH, LOC=\"{}\" *)".format(site_name))
            print(
                """GTPE2_CHANNEL {attrs} {site} (
    {ports}
);
            """.format(
                    attrs=verilog_attr,
                    site=tile_type.lower(),
                    ports=verilog_ports.rstrip(",")))

        params_list.append(params)
        params_dict["params"] = params_list
        primitives_list.append(params_dict)

    for l in luts.create_wires_and_luts():
        print(l)

    print("endmodule")

    with open('params.json', 'w') as f:
        json.dump(primitives_list, f, indent=2)


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2020  The Project X-Ray Authors.
#
# Use of this source code is governed by a ISC-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/ISC
#
# SPDX-License-Identifier: ISC
import json
import os
from prjxray.util import block_type_s2i

# Break frames into WORD_SIZE bit words.
WORD_SIZE_BITS = 32

# How many 32-bit words for frame in a 7-series bitstream?
FRAME_WORD_COUNT = 101

# What alignment is expect for columns.
FRAME_ALIGNMENT = 0x80
'''
Sample:
bit_0002000f_079_06

Where:
-0002000f: FDRI address
-079: FDIR word number (0-100)
-06: bit index (0-31)
'''


def load_bitdata(f):
    """ Read bit file and return bitdata map.
    Similar to segbits file

    bitdata is a map of of two sets.
    The map key is the frame address.
    The first sets are the word columns that have any bits set.
    Word columsn are WORD_SIZE_BITS wide.
    The second sets are bit index within the frame and word if it is set.
    """
    bitdata = dict()

    for line in f:
        line = line.split("_")
        frame = int(line[1], 16)
        wordidx = int(line[2], 10)
        bitidx = int(line[3], 10)

        if frame not in bitdata:
            bitdata[frame] = set(), set()

        bitdata[frame][0].add(wordidx)
        bitdata[frame][1].add(wordidx * WORD_SIZE_BITS + bitidx)

    return bitdata


# used by segprint
# TODO: merge these
def load_bitdata2(f):
    '''
    return as bitdata[frame][wordidx].add(bitidx)
    ie indexed by frame, word index, and then a set with bit indexes
    Similar to .bits file: bit_00020012_014_20
    '''

    bitdata = dict()

    for lineraw in f:
        lineraw = lineraw.strip()
        line = lineraw.split("_")
        try:
            frame = int(line[1], 16)
            wordidx = int(line[2], 10)
            bitidx = int(line[3], 10)
        except:
            print("Invalid line %s" % lineraw)
            raise

        if frame not in bitdata:
            bitdata[frame] = dict()

        if wordidx not in bitdata[frame]:
            bitdata[frame][wordidx] = set()

        bitdata[frame][wordidx].add(bitidx)
    return bitdata


def gen_part_base_addrs():
    """
    Return (block_type, top_bottom, cfg_row, cfg_col, frame_count)
    Where:
    -block_type ("bus"): typically CLB_IO_CLK, sometimes BLOCK_RAM
    -top_bottom: either "top" or "bottom"
    -cfg_row: a relative row
    -cfg_col: a relative column
    -frame_count: number of frames to fully configure this minor address

    Example:
    ('CLB_IO_CLK', 'bottom', 0, 3, 36)
    ('BLOCK_RAM', 'top', 0, 1, 128)
    ('CLB_IO_CLK', 'top', 1, 34, 28)
    """
    fn = os.getenv("XRAY_PART_YAML").replace(".yaml", ".json")
    j = json.load(open(fn, "r"))
    for tbk, tbv in j["global_clock_regions"].items():
        for rowk, rowv in tbv["rows"].items():
            for busk, busv in rowv["configuration_buses"].items():
                for colk, colv in busv["configuration_columns"].items():
                    yield (
                        busk, tbk, int(rowk), int(colk), colv["frame_count"])


def addr_bits2word(block_type, top_bottom, cfg_row, cfg_col, minor_addr):
    """Convert a deconstructed address to a 32 bit word"""
    # https://www.xilinx.com/support/documentation/user_guides/ug470_7Series_Config.pdf
    ret = 0
    ret |= block_type_s2i[block_type] << 23
    ret |= {"top": 0, "bottom": 1}[top_bottom] << 22
    ret |= cfg_row << 17
    ret |= cfg_col << 7
    ret |= minor_addr
    return ret



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2020  The Project X-Ray Authors.
#
# Use of this source code is governed by a ISC-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/ISC
#
# SPDX-License-Identifier: ISC
dfii_control_sel = 0x01
dfii_control_cke = 0x02
dfii_control_odt = 0x04
dfii_control_reset_n = 0x08

dfii_command_cs = 0x01
dfii_command_we = 0x02
dfii_command_cas = 0x04
dfii_command_ras = 0x08
dfii_command_wrdata = 0x10
dfii_command_rddata = 0x20

ddrx_mr1 = 0x6

init_sequence = [
    ("Release reset", 0, 0, dfii_control_odt | dfii_control_reset_n, 50000),
    (
        "Bring CKE high", 0, 0,
        dfii_control_cke | dfii_control_odt | dfii_control_reset_n, 10000),
    (
        "Load Mode Register 2, CWL=5", 512, 2, dfii_command_ras
        | dfii_command_cas | dfii_command_we | dfii_command_cs, 0),
    (
        "Load Mode Register 3", 0, 3, dfii_command_ras | dfii_command_cas
        | dfii_command_we | dfii_command_cs, 0),
    (
        "Load Mode Register 1", 6, 1, dfii_command_ras | dfii_command_cas
        | dfii_command_we | dfii_command_cs, 0),
    (
        "Load Mode Register 0, CL=6, BL=8", 2336, 0, dfii_command_ras
        | dfii_command_cas | dfii_command_we | dfii_command_cs, 200),
    ("ZQ Calibration", 1024, 0, dfii_command_we | dfii_command_cs, 200),
]



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2020  The Project X-Ray Authors.
#
# Use of this source code is governed by a ISC-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/ISC
#
# SPDX-License-Identifier: ISC
""" Tool for building zero db file for BUFG pips.
This requires that the rdb files be good enough to identify all the 0 candidate
features, which may take multiple manual iterations.  Manual iterations can
be running like:

make ITER=<N> -j<J> database

And then invoking:
python3 build_zdb.py build/segbits_clk_bufg_bot_r.rdb build/segbits_clk_bufg_top_r.rdb > bits.dbf

will successed if and only if the rdb is complete enough.

bits.dbf is committed, so this utility should only be needed to document the
process.

"""
import argparse


def main():
    parser = argparse.ArgumentParser("Form ZDB groups for BUFG.")

    parser.add_argument('bot_r')
    parser.add_argument('top_r')

    args = parser.parse_args()

    groups = {}

    with open(args.bot_r) as f:
        for l in f:
            parts = l.strip().split(' ')
            feature = parts[0]
            bits = parts[1:]
            tile_type, dst, src = feature.split('.')

            assert tile_type == "CLK_BUFG"

            if dst not in groups:
                groups[dst] = {}

            groups[dst][src] = bits

    print('# Generated from build_zdb.py')

    for dst in groups:
        if len(groups[dst]) == 1:
            continue

        bits = set()
        zero_feature = None
        for src in groups[dst]:
            if groups[dst][src][0] == '<0':
                assert zero_feature is None
                zero_feature = src
            else:
                bits |= set(groups[dst][src])

        assert zero_feature is not None, dst

        print(
            '{bits},{type}.{dst}.{src}'.format(
                bits=' '.join(sorted(bits)),
                type='CLK_BUFG',
                dst=dst,
                src=zero_feature))


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2020  The Project X-Ray Authors.
#
# Use of this source code is governed by a ISC-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/ISC
#
# SPDX-License-Identifier: ISC
import argparse

REGISTER_LAYOUT = {
    'CLKOUT1': [
        ('LOW_TIME', 6),
        ('HIGH_TIME', 6),
        ('OUTPUT_ENABLE', 1),
        ('PHASE_MUX', 3),
    ],
    'CLKOUT2': [
        ('DELAY_TIME', 6),
        ('NO_COUNT', 1),
        ('EDGE', 1),
        ('MX', 2),
        ('FRAC_WF_R', 1),
        ('FRAC_EN', 1),
        ('FRAC', 3),
        ('RESERVED', 1),
    ],
    'DIVCLK': [
        ('LOW_TIME', 6),
        ('HIGH_TIME', 6),
        ('NO_COUNT', 1),
        ('EDGE', 1),
        ('RESERVED', 2),
    ],
    'LOCKREG1': [
        ('LKTABLE', 10, 20),
        ('LOCKREG1_RESERVED', 6, 0),
    ],
    'LOCKREG2': [
        ('LKTABLE', 10, 0),
        ('LKTABLE', 5, 30),
        ('LOCKREG2_RESERVED', 1, 0),
    ],
    'LOCKREG3': [
        ('LKTABLE', 10, 10),
        ('LKTABLE', 5, 35),
        ('LOCKREG3_RESERVED', 1, 0),
    ],
    'FILTREG1': [
        ('FILTREG1_RESERVED', 8, 0),
        ('TABLE', 1, 6),
        ('FILTREG1_RESERVED', 2, 8),
        ('TABLE', 2, 7),
        ('FILTREG1_RESERVED', 2, 10),
        ('TABLE', 1, 9),
    ],
    'FILTREG2': [
        ('FILTREG2_RESERVED', 4, 0),
        ('TABLE', 1, 0),
        ('FILTREG2_RESERVED', 2, 4),
        ('TABLE', 2, 1),
        ('FILTREG2_RESERVED', 2, 6),
        ('TABLE', 2, 3),
        ('FILTREG2_RESERVED', 2, 8),
        ('TABLE', 1, 5),
    ],
    'POWER_REG': [
        ('POWER_REG', 16),
    ],
}

BASE_OFFSET = 0x00
REGISTER_MAP = []

REGISTER_MAP.append(None)
REGISTER_MAP.append(None)

for idx in range(3):
    REGISTER_MAP.append(None)

REGISTER_MAP.append(None)

# 0x06 - 0x15
for output in ['CLKOUT5', 'CLKOUT0', 'CLKOUT1', 'CLKOUT2', 'CLKOUT3',
               'CLKOUT4', None, 'CLKFBOUT']:
    if output is not None:
        REGISTER_MAP.append(('CLKOUT1', output))
        REGISTER_MAP.append(('CLKOUT2', output))
    else:
        REGISTER_MAP.append(None)
        REGISTER_MAP.append(None)

# 0x16
REGISTER_MAP.append(('DIVCLK', 'DIVCLK'))
# 0x17
REGISTER_MAP.append(None)
# 0x18-0x1A
REGISTER_MAP.append(('LOCKREG1', 'LOCKREG1'))
REGISTER_MAP.append(('LOCKREG2', 'LOCKREG2'))
REGISTER_MAP.append(('LOCKREG3', 'LOCKREG3'))

for _ in range(0x28 - 0x1A - 1):
    REGISTER_MAP.append(None)

REGISTER_MAP.append(('POWER_REG', 'POWER_REG'))

for _ in range(0x4E - 0x28 - 1):
    REGISTER_MAP.append(None)

# 0x4E - 0x4F
REGISTER_MAP.append(('FILTREG1', 'FILTREG1'))
REGISTER_MAP.append(('FILTREG2', 'FILTREG2'))

for _ in range(0x20):
    REGISTER_MAP.append(None)


class RegisterAddress(object):
    def __init__(self, frame_offsets, bit_offset):
        self.frame_index = 0
        self.frame_offsets = frame_offsets
        self.bit_offset = bit_offset
        self.bits_used = set()

    def next_bit(self, used=True):
        output = '{}_{}'.format(
            self.frame_offsets[self.frame_index], self.bit_offset)

        if used:
            self.bits_used.add(output)

        self.frame_index += 1
        if self.frame_index >= len(self.frame_offsets):
            self.frame_index = 0
            self.bit_offset += 1

        return output


def passthrough_non_register_segbits(seg_in):
    """ Filter input segbits file and capture register base offset.

    Some PLL bit ranges are documented registers in the PLL/MMCM dynamic
    reconfiguration iterface.  These features will be generated in
    output_registers.  In order for output_registers to function, it needs
    the starting bit offset of the register space, which is based off of
    base_offset_register segbit definition.

    Other features generated in fuzzing are passed through.

    """
    base_offset_register = 'CMT_UPPER_T.PLLE2_ADV.CLKOUT5_DIVIDE[1]'

    bit_offset = None
    in_use = None
    with open(seg_in, 'r') as f:
        for l in f:
            if l.startswith(base_offset_register):
                parts = l.split()
                assert len(parts) == 2
                assert parts[0] == base_offset_register
                frame_offset, bit_index = map(int, parts[1].split('_'))

                assert frame_offset == 28
                assert bit_index > 3
                bit_offset = bit_index - 3 - 16 * 3

                continue

            if 'IN_USE' in l:
                assert in_use is None
                in_use = l.strip()
                continue

            parts = l.split()
            feature_parts = parts[0].split('.')

            if len(feature_parts) < 3:
                print(l.strip())
                continue

            if feature_parts[2] == 'BANDWIDTH':
                continue

            if '[' not in feature_parts[2]:
                print(l.strip())
                continue

            base_feature = feature_parts[2].split('[')

            if base_feature[0] in [
                    'CLKOUT0_DIVIDE',
                    'CLKOUT1_DIVIDE',
                    'CLKOUT2_DIVIDE',
                    'CLKOUT3_DIVIDE',
                    'CLKOUT4_DIVIDE',
                    'CLKOUT5_DIVIDE',
                    'DIVCLK_DIVIDE',
                    'CLKFBOUT_MULT',
                    'CLKOUT0_DUTY_CYCLE',
            ]:
                # These features are PLL registers, so ignore the base
                continue

            print(l.strip())

    assert bit_offset is not None
    assert in_use is not None
    return bit_offset, in_use


def output_registers(bit_offset, in_use):
    """ Output segbits for the known PLL register space.

    The first bit offset in the register space is required to generate this
    output.

    """
    reg = RegisterAddress(frame_offsets=[28, 29], bit_offset=bit_offset)

    for idx, register in enumerate(REGISTER_MAP):
        if register is None:
            for _ in range(16):
                reg.next_bit(used=False)
            continue

        layout, register_name = register

        layout_bits = REGISTER_LAYOUT[layout]

        simple_layout = len(layout_bits[0]) == 2

        if True:
            bit_count = 0
            if simple_layout:
                for field, width in layout_bits:
                    for bit in range(width):
                        bit_count += 1

                        if field is None:
                            reg.next_bit(used=False)
                            continue

                        print(
                            'CMT_UPPER_T.PLLE2_ADV.{}_{}_{}[{}] {}'.format(
                                register_name, layout, field, bit,
                                reg.next_bit()))
            else:
                for field, width, start_bit in layout_bits:
                    for bit in range(width):
                        bit_count += 1

                        if field is None:
                            reg.next_bit(used=False)
                            continue

                        print(
                            'CMT_UPPER_T.PLLE2_ADV.{}[{}] {}'.format(
                                field, start_bit + bit, reg.next_bit()))

            assert bit_count == 16
        else:
            for bit in range(16):
                if register_name != layout or layout in ['CLKOUT1', 'CLKOUT2']:
                    print(
                        'CMT_UPPER_T.PLLE2_ADV.{}_{}[{}] {}'.format(
                            register_name, layout, bit, reg.next_bit()))
                else:
                    print(
                        'CMT_UPPER_T.PLLE2_ADV.{}[{}] {}'.format(
                            register_name, bit, reg.next_bit()))

    parts = in_use.split()
    feature = parts[0]
    bits = [p for p in parts[1:] if p not in reg.bits_used]
    print('{} {}'.format(feature, ' '.join(bits)))


def main():
    parser = argparse.ArgumentParser(description="")

    parser.add_argument('--seg_in')

    args = parser.parse_args()

    bit_offset, in_use = passthrough_non_register_segbits(args.seg_in)

    output_registers(bit_offset, in_use)


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
# Copyright (C) 2017-2022 The Project X-Ray Authors.
#
# Use of this source code is governed by a ISC-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/ISC
#
# SPDX-License-Identifier: ISC
#
# Project X-Ray documentation build configuration file, created by
# sphinx-quickstart on Mon Feb  5 11:04:37 2018.
#
# This file is execfile()d with the current directory set to its
# containing dir.
#
# Note that not all possible configuration values are present in this
# autogenerated file.
#
# All configuration values have a default; values that are commented out
# serve to show the default.

# Updated documentation of the configuration options is available at
# https://www.sphinx-doc.org/en/master/usage/configuration.html

from pathlib import Path
import re
import recommonmark
import os
from sys import path as sys_path
import subprocess

sys_path.insert(0, os.path.abspath('.'))
from markdown_code_symlinks import LinkParser, MarkdownSymlinksDomain

# -- General configuration ------------------------------------------------

project = 'Project X-Ray'
author = 'F4PGA Authors'
copyright = f'{author}, 2018 - 2022'

# yapf: disable
extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.autosummary',
    'sphinx.ext.doctest',
    'sphinx.ext.imgmath',
    'sphinx.ext.napoleon',
    'sphinx.ext.todo',
    'sphinx_markdown_tables',
    'recommonmark'
]
# yapf: enable

emplates_path = ['_templates']

source_suffix = ['.rst', '.md']
source_parsers = {
    '.md': 'markdown_code_symlinks.LinkParser',
}

master_doc = 'index'

# Enable github links when not on readthedocs
on_rtd = os.environ.get('READTHEDOCS', None) == 'True'
if not on_rtd:
    html_context = {
        "display_github": True,  # Integrate GitHub
        "github_user": "f4pga",  # Username
        "github_repo": "prjxray",  # Repo name
        "github_version": "master",  # Version
        "conf_py_path": "/doc/",
    }
else:
    docs_dir = os.path.abspath(os.path.dirname(__file__))
    print("Docs dir is:", docs_dir)
    subprocess.call('git fetch origin --unshallow', cwd=docs_dir, shell=True)
    subprocess.check_call('git fetch origin --tags', cwd=docs_dir, shell=True)
    subprocess.check_call('make links', cwd=docs_dir, shell=True)

# The full version, including alpha/beta/rc tags.
release = re.sub('^v', '', os.popen('git describe ').read().strip())
# The short X.Y version.
version = release

# yapf: disable
exclude_patterns = [
    '_build',
    'architecture/copying.md',
    'db_dev_process/minitests/index/**',
    'db_dev_process/fuzzers/index/**'
]
# yapf: enable

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'default'

# If true, `todo` and `todoList` produce output, else they produce nothing.
todo_include_todos = True

# -- Options for HTML output ----------------------------------------------

html_show_sourcelink = True

html_theme = 'sphinx_symbiflow_theme'

html_theme_options = {
    'repo_name': 'f4pga/prjxray',
    'github_url' : 'https://github.com/f4pga/prjxray',
    'globaltoc_collapse': True,
    'color_primary': 'indigo',
    'color_accent': 'blue',
}

html_static_path = ['_static']

html_logo = str(Path(html_static_path[0]) / 'logo.svg')
html_favicon = str(Path(html_static_path[0]) / 'favicon.svg')

# -- Options for HTMLHelp output ------------------------------------------

# Output file base name for HTML help builder.
htmlhelp_basename = 'prjxray'

# -- Options for LaTeX output ---------------------------------------------

latex_elements = {}

latex_documents = [
    (
        master_doc, 'ProjectX-Ray.tex', u'Project X-Ray Documentation',
        u'F4PGA Authors', 'manual'),
]

# -- Options for manual page output ---------------------------------------

man_pages = [
    (master_doc, 'projectx-ray', u'Project X-Ray Documentation', [author], 1)
]

# -- Options for Texinfo output -------------------------------------------

texinfo_documents = [
    (
        master_doc, 'ProjectX-Ray', u'Project X-Ray Documentation', author,
        'ProjectX-Ray', 'One line description of project.', 'Miscellaneous'),
]

intersphinx_mapping = {'https://docs.python.org/': None}


def setup(app):
    # Generate links for markdown-code-symlinks
    subprocess.check_call("make links", shell=True)

    github_code_repo = 'https://github.com/f4pga/prjxray/'
    github_code_branch = 'blob/master/'

    docs_root_dir = os.path.realpath(os.path.dirname(__file__))
    code_root_dir = os.path.realpath(os.path.join(docs_root_dir, ".."))

    MarkdownSymlinksDomain.init_domain(
        github_code_repo, github_code_branch, docs_root_dir, code_root_dir)
    MarkdownSymlinksDomain.find_links()
    app.add_domain(MarkdownSymlinksDomain)
    app.add_config_value(
        'recommonmark_config', {
            'github_code_repo': github_code_repo,
        }, True)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2020  The Project X-Ray Authors.
#
# Use of this source code is governed by a ISC-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/ISC
#
# SPDX-License-Identifier: ISC

import os, json, re

#################################################
# Loading Raw Source Data

grid = None
bits = dict()
luts = dict()
carry = dict()

print("Loading grid.")
with open("../../../gridinfo/grid-%s-db.json" % os.getenv("XRAY_PART"),
          "r") as f:
    grid = json.load(f)

print("Loading bits.")
with open("design.bits", "r") as f:
    for line in f:
        line = line.split("_")
        bit_frame = int(line[1], 16)
        bit_wordidx = int(line[2], 16)
        bit_bitidx = int(line[3], 16)
        base_frame = bit_frame & ~0x7f

        if base_frame not in bits:
            bits[base_frame] = dict()

        if bit_wordidx not in bits[base_frame]:
            bits[base_frame][bit_wordidx] = set()

        bits[base_frame][bit_wordidx].add((bit_frame, bit_wordidx, bit_bitidx))

print("Loading lut data.")
with open("lutdata.txt", "r") as f:
    for line in f:
        line = line.split()
        site = line[0]
        bel = line[1]
        init = int(line[2][4:], 16)

        if site not in luts:
            luts[site] = dict()

        for i in range(64):
            bitname = "%s.INIT[%02d]" % (bel, i)
            luts[site][bitname] = ((init >> i) & 1) != 0

print("Loading carry data.")
with open("carrydata.txt", "r") as f:
    for line in f:
        line = line.split()
        assert line[0] not in carry
        carry[line[0]] = dict()
        for i, n in enumerate(
                "CYINIT:ZRO:ONE:AX:CIN DI0:AX:O5 DI1:AX:O5 DI2:AX:O5 DI3:AX:O5"
                .split()):
            n = n.split(":")
            for k in n[1:]:
                carry[line[0]]["CARRY_%s_MUX_%s" %
                               (n[0], k)] = line[1 + i].upper() == k

#################################################
# Group per Segment

print("Compile segment data.")

segments = dict()

for tilename, tiledata in grid["tiles"].items():
    found_data = False
    for site in tiledata["sites"]:
        if site in luts:
            found_data = True

    if not found_data:
        continue

    segname = "%s_%02x" % (
        tiledata["cfgcol"]["BASE_FRAMEID"][2:], min(
            tiledata["cfgcol"]["WORDS"]))

    if not segname in segments:
        segments[segname] = {"bits": list(), "tags": dict()}

    for site in tiledata["sites"]:
        if site not in luts and site not in carry:
            continue

        if re.match(r"SLICE_X[0-9]*[02468]Y", site):
            sitekey = "SLICE_X0"
        elif re.match(r"SLICE_X[0-9]*[13579]Y", site):
            sitekey = "SLICE_X1"
        else:
            assert 0

        if site in luts:
            for name, value in luts[site].items():
                tile_type = tiledata["props"]["TYPE"]

                # LUT init bits are in the same position for all CLBL[LM]_[LR] tiles
                if re.match("^CLBL[LM]_[LR]",
                            tile_type) and "LUT.INIT" in name:
                    tile_type = "CLBLX_X"

                segments[segname]["tags"]["%s.%s.%s" %
                                          (tile_type, sitekey, name)] = value

        if site in carry:
            for name, value in carry[site].items():
                tile_type = tiledata["props"]["TYPE"]
                segments[segname]["tags"]["%s.%s.%s" %
                                          (tile_type, sitekey, name)] = value

    base_frame = int(tiledata["cfgcol"]["BASE_FRAMEID"][2:], 16)
    for wordidx in tiledata["cfgcol"]["WORDS"]:
        if base_frame not in bits:
            continue
        if wordidx not in bits[base_frame]:
            continue
        for bit_frame, bit_wordidx, bit_bitidx in bits[base_frame][wordidx]:
            segments[segname]["bits"].append(
                "%02x_%02x_%02x" % (
                    bit_frame - base_frame, bit_wordidx - min(
                        tiledata["cfgcol"]["WORDS"]), bit_bitidx))

    segments[segname]["bits"].sort()

#################################################
# Print

print("Write segment data.")

with open("segdata.txt", "w") as f:
    for segname, segdata in sorted(segments.items()):
        print("seg %s" % segname, file=f)
        for bitname in sorted(segdata["bits"]):
            print("bit %s" % bitname, file=f)
        for tagname, tagval in sorted(segdata["tags"].items()):
            print("tag %s %d" % (tagname, tagval), file=f)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2020  The Project X-Ray Authors.
#
# Use of this source code is governed by a ISC-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/ISC
#
# SPDX-License-Identifier: ISC
import os
import random
random.seed(int(os.getenv("SEED"), 16))
from prjxray import util
from prjxray.db import Database


def gen_sites():
    db = Database(util.get_db_root(), util.get_part())
    grid = db.grid()
    for tile_name in sorted(grid.tiles()):
        loc = grid.loc_of_tilename(tile_name)
        gridinfo = grid.gridinfo_at_loc(loc)

        found_xadc = False
        for site_name, site_type in gridinfo.sites.items():
            if site_type in ['XADC']:
                found_xadc = True
                break

        if not found_xadc:
            continue

        int_grid_x = loc.grid_x + 3
        int_tile_type = 'INT_L'

        int_tile_locs = []
        for dy in range(-1, 1):
            int_tile_locs.append((int_grid_x, loc.grid_y + dy), )

        int_tiles = []
        for int_tile_loc in int_tile_locs:
            int_gridinfo = grid.gridinfo_at_loc(int_tile_loc)
            assert int_gridinfo.tile_type == int_tile_type, (
                int_gridinfo.tile_type, int_tile_type)

            int_tiles.append(grid.tilename_at_loc(int_tile_loc))

        yield tile_name, int_tiles


def write_params(params):
    pinstr = 'tile,val\n'
    for tile, (val) in sorted(params.items()):
        pinstr += '%s,%s\n' % (tile, val)
    open('params.csv', 'w').write(pinstr)


def run():
    print('''
module top();
    ''')

    sites = list(gen_sites())

    # Only on CFG_CENTER_MID expected.
    assert len(sites) == 1
    tile_name, int_tiles = sites[0]

    # int_tiles[0]:
    # IMUX43 -> XADC_CONVST = 0
    # IMUX42 -> XADC_DWE = toggle 0/1
    # int_tiles[1]:
    # IMUX43 -> XADC_DI15 = 0
    # IMUX42 -> XADC_DI14 = toggle 0/1

    DWE = random.randint(0, 1)
    DI14 = random.randint(0, 1)

    params = {}
    params[int_tiles[0]] = DWE
    params[int_tiles[1]] = DI14

    print(
        """
    wire [15:0] di;
    wire dwe;
    wire convst;

    assign convst = 0;
    assign dwe = {DWE};

    assign di[15] = 0;
    assign di[14] = {DI14};

    (* KEEP, DONT_TOUCH *)
    XADC xadc (
        .DI(di),
        .DWE(dwe),
        .CONVST(convst)
    );
    """.format(
            DWE=DWE,
            DI14=DI14,
        ))

    print("endmodule")
    write_params(params)


if __name__ == '__main__':
    run()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2020  The Project X-Ray Authors.
#
# Use of this source code is governed by a ISC-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/ISC
#
# SPDX-License-Identifier: ISC

import unittest
import demo_sw_led


class TestStringMethods(unittest.TestCase):
    def test_all(self):
        for i in range(8):
            print()
            print()
            print()
            print(i)
            demo_sw_led.run('out_xc7a35tcpg236-1_BASYS3-SWBUT_roi_basev', i, i)


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2020  The Project X-Ray Authors.
#
# Use of this source code is governed by a ISC-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/ISC
#
# SPDX-License-Identifier: ISC
"""
This script receives and parses data generated by the "histogram" test design.
"""

import argparse
import serial

# =============================================================================


def main():

    parser = argparse.ArgumentParser(
        description=__doc__,
        formatter_class=argparse.RawDescriptionHelpFormatter)
    parser.add_argument("port", type=str, help="Serial port name")
    parser.add_argument("--baud", type=int, default=115200, help="Baudrate")

    args = parser.parse_args()

    # Open serial port
    port = serial.Serial(args.port, baudrate=args.baud)

    # Get first line and discard it. It may be broken
    port.readline()

    # Read and process lines
    while True:

        line = port.readline()
        line = line.decode("ASCII").strip()

        data = [int(x, base=16) for x in line.split("_")]
        print(" ".join("%4d" % x for x in data))


# =============================================================================

if __name__ == "__main__":
    main()
